from unittest import TestCase
import odmlib.define_2_1.model as DEFINE
import odmlib.odm_parser as ODM_PARSER
import xml.etree.ElementTree as ET
import odmlib.ns_registry as NS
import odmlib.define_loader as OL
import odmlib.loader as LD
import os
import datetime


class TestItemGroupDef(TestCase):
    def setUp(self) -> None:
        attrs = self.set_itemgroupdef_attributes()
        self.igd = DEFINE.ItemGroupDef(**attrs)
        self.test_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'defineV21-SDTM-test.xml')
        self.test_file_json = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'defineV21-SDTM-test.json')
        self.input_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'defineV21-SDTM.xml')
        self.nsr = NS.NamespaceRegistry(prefix="odm", uri="http://www.cdisc.org/ns/odm/v1.3", is_default=True)
        self.nsr = NS.NamespaceRegistry(prefix="def", uri="http://www.cdisc.org/ns/def/v2.1")

    def test_context(self):
        odm = DEFINE.ODM(
            FileOID="DEFINE.TEST.IGD.001", FileType="Snapshot", AsOfDateTime=self.set_datetime(),
            CreationDateTime=self.set_datetime(), ODMVersion="1.3.2", Originator="CDISC 360", SourceSystem="WS2",
            SourceSystemVersion="0.1", Context="Other"
        )
        self.assertEqual(odm.Context, "Other")

    def test_standards(self):
        mdv = self.create_mdv()
        mdv.Standards = DEFINE.Standards()
        std1 = DEFINE.Standard(OID="STD.1", Name="SDTMIG", Type="IG", Version="3.1.2", Status="Final", CommentOID="COM.STD1")
        mdv.Standards.Standard.append(std1)
        std2 = DEFINE.Standard(OID="STD.2", Name="CDISC/NCI", Type="CT", PublishingSet="SDTM", Version="2011-12-09", Status="Final")
        mdv.Standards.Standard.append(std2)
        self.assertEqual(mdv.Standards.Standard[0].Name, "SDTMIG")
        self.assertEqual(mdv.Standards.Standard[1].Version, "2011-12-09")

    def test_add_class(self):
        def_class = DEFINE.Class(Name="FINDINGS")
        def_class.SubClass.append(DEFINE.SubClass(Name="SUBFINDINGS", ParentClass="FINDINGS"))
        self.igd.Class = def_class
        self.assertEqual(self.igd.Class.Name, "FINDINGS")
        self.assertEqual(self.igd.Class.SubClass[0].ParentClass, "FINDINGS")

    def test_add_description(self):
        tt1 = DEFINE.TranslatedText(_content="this is the first test description", lang="en")
        tt2 = DEFINE.TranslatedText(_content="this is the second test description", lang="en")
        self.igd.Description = DEFINE.Description()
        self.igd.Description.TranslatedText = [tt1, tt2]
        self.assertEqual(len(self.igd.Description.TranslatedText), 2)
        self.assertEqual(self.igd.Description.TranslatedText[1]._content, 'this is the second test description')

    def test_add_item_ref(self):
        self.igd.ItemRef.append(DEFINE.ItemRef(ItemOID="IT.STUDYID", Mandatory="Yes", OrderNumber=1, IsNonStandard="Yes"))
        self.igd.ItemRef.append(DEFINE.ItemRef(ItemOID="IT.VS.VSTEST", Mandatory="No", OrderNumber=2, HasNoData="Yes"))
        self.igd.ItemRef.append(DEFINE.ItemRef(ItemOID="IT.VS.VSORRES", Mandatory="Yes", OrderNumber=3, MethodOID="MT.METHODFEX"))
        self.assertEqual(self.igd.ItemRef[0].IsNonStandard, "Yes")
        self.assertEqual(self.igd.ItemRef[1].HasNoData, "Yes")

    def test_add_item_ref_list(self):
        ir1 = DEFINE.ItemRef(ItemOID="IT.STUDYID", Mandatory="Yes", OrderNumber=1)
        self.igd.ItemRef.append(ir1)
        ir2 = DEFINE.ItemRef(ItemOID="IT.VS.VSTEST", Mandatory="No", OrderNumber=2)
        self.igd.ItemRef.append(ir2)
        ir3 = DEFINE.ItemRef(ItemOID="IT.VS.VSORRES", Mandatory="Yes", OrderNumber=3, MethodOID="MT.METHODFEX")
        self.igd.ItemRef.append(ir3)
        self.assertEqual(self.igd.ItemRef[0].ItemOID, "IT.STUDYID")
        self.assertEqual(self.igd.ItemRef[2].MethodOID, "MT.METHODFEX")

    def test_add_item_ref_missing_kwarg(self):
        with self.assertRaises(ValueError):
            self.igd.ItemRef = [DEFINE.ItemRef(Mandatory="Yes", OrderNumber=1)]

    def test_add_item_ref_invalid_kwarg(self):
        with self.assertRaises(TypeError):
            self.igd.ItemRef = [DEFINE.ItemRef(ItemOID="IT.STUDYID", Mandatory="Yes", InValid="Yes")]

    def test_item_ref_exists(self):
        self.igd.ItemRef = [DEFINE.ItemRef(ItemOID="IT.VS.VSTESTCD", Mandatory="Yes", OrderNumber=4)]
        self.assertEqual(self.igd.ItemRef[0].ItemOID, "IT.VS.VSTESTCD")

    def test_add_alias(self):
        a1 = DEFINE.Alias(Context="SDTMIG", Name="VSORRES")
        a2 = DEFINE.Alias(Context="SDTMIG", Name="VSTESTCD")
        self.igd.Alias = [a1, a2]
        self.assertEqual(len(self.igd.Alias), 2)
        self.assertEqual(self.igd.Alias[1].Name, "VSTESTCD")

    def test_to_xml(self):
        attrs = self.set_itemgroupdef_attributes()
        igd = DEFINE.ItemGroupDef(**attrs)
        tt = DEFINE.TranslatedText(_content="this is the first test description", lang="en")
        desc = DEFINE.Description()
        desc.TranslatedText = [tt]
        igd.Description = desc
        ir1 = DEFINE.ItemRef(ItemOID="IT.STUDYID", Mandatory="Yes", OrderNumber=1)
        ir2 = DEFINE.ItemRef(ItemOID="IT.VS.VSTEST", Mandatory="No", OrderNumber=2)
        igd.ItemRef = [ir1, ir2]
        igd_xml = igd.to_xml()
        self.assertEqual(igd_xml.attrib["OID"], "IG.VS")
        self.assertListEqual(["Description", "ItemRef", "ItemRef"], [e.tag for e in igd_xml.getchildren()])

    def test_itemgroupdef_parse_xml(self):
        parser = ODM_PARSER.ODMParser(self.input_file, self.nsr)
        parser.parse()
        mdv = parser.MetaDataVersion()
        igd_list = parser.ItemGroupDef(parent=mdv[0])
        igd_dict = igd_list[0]
        self.assertEqual(igd_dict["OID"], "IG.TS")
        ir_list = parser.ItemRef(parent=igd_dict["elem"])
        # tests the __len__ in ItemGroupDef as well as the add_item_ref
        self.assertEqual(len(ir_list), 6)
        igd_class = parser.Class(parent=igd_dict["elem"], ns_prefix="def")
        self.assertEqual(igd_class[0]["Name"], "TRIAL DESIGN")

    def test_write_xml(self):
        attrs = self.set_itemgroupdef_attributes()
        igd = DEFINE.ItemGroupDef(**attrs)
        tt = DEFINE.TranslatedText(_content="this is the first test description", lang="en")
        desc = DEFINE.Description()
        desc.TranslatedText = [tt]
        igd.Description = desc
        ir_list = self.set_itemrefs()
        igd.ItemRef = ir_list
        self.create_odm_document(igd)
        loader = LD.ODMLoader(OL.XMLDefineLoader(model_package="define_2_1", ns_uri="http://www.cdisc.org/ns/def/v2.1"))
        loader.open_odm_document(self.test_file)
        mdv = loader.MetaDataVersion()

        igd_list = mdv.ItemGroupDef
        igd = igd_list[0]
        self.assertEqual(igd.OID, "IG.VS")
        # tests the __len__ in ItemGroupDef
        self.assertEqual(len(igd.ItemRef), 10)

    def test_write_json(self):
        attrs = self.set_itemgroupdef_attributes()
        igd = DEFINE.ItemGroupDef(**attrs)
        tt = DEFINE.TranslatedText(_content="this is the first test description", lang="en")
        desc = DEFINE.Description()
        desc.TranslatedText = [tt]
        igd.Description = desc
        ir_list = self.set_itemrefs()
        igd.ItemRef = ir_list
        self.create_odm_document_json(igd)
        loader = LD.ODMLoader(OL.JSONDefineLoader(model_package="define_2_1"))
        loader.open_odm_document(self.test_file_json)
        mdv = loader.MetaDataVersion()

        igd_list = mdv.ItemGroupDef
        igd = igd_list[0]
        self.assertEqual(igd.OID, "IG.VS")
        # tests the __len__ in ItemGroupDef
        self.assertEqual(len(igd.ItemRef), 10)

    @staticmethod
    def add_itemrefs(igd, item_refs):
        """
        add ItemRefs to the ItemGroupDef object
        :param igd: ItemGroupDef object
        :param item_refs: list of ItemRef dictionaries containing ItemRef attributes
        """
        for it in item_refs:
            attrs = {"ItemOID": it.oid, "Mandatory": it.mandatory, "OrderNumber": it.order_number}
            if it.key_sequence:
                attrs["KeySequence"] = it.key_sequence
            if it.method:
                attrs["MethodOID"] = it.method
            igd.add_item_ref(**attrs)

    @staticmethod
    def set_itemgroupdef_attributes():
        """
        set some ItemGroupDef element attributes using test data
        :return: dictionary with ItemGroupDef attribute information
        """
        return {"OID": "IG.VS", "Name": "VS", "Repeating": "Yes", "Domain": "VS", "SASDatasetName": "VS",
                "IsReferenceData": "No", "Purpose": "Tabulation", "ArchiveLocationID": "LF.VS",
                "Structure": "One record per vital sign measurement per visit per subject", "StandardOID": "STD.1",
                "IsNonStandard": "Yes", "HasNoData": "Yes"}

    @staticmethod
    def set_itemrefs():
        """
        set some ItemRef element attributes using test data
        :return: return a list of ItemRef named tuples
        """
        itemrefs = [
            DEFINE.ItemRef(ItemOID="IT.STUDYID", Mandatory="Yes", OrderNumber=1, KeySequence=1),
            DEFINE.ItemRef(ItemOID="IT.TA.DOMAIN", Mandatory="Yes", OrderNumber=2),
            DEFINE.ItemRef(ItemOID="IT.TA.ARMCD", Mandatory="Yes", OrderNumber=3, KeySequence=2),
            DEFINE.ItemRef(ItemOID="IT.TA.ARM", Mandatory="Yes", OrderNumber=4),
            DEFINE.ItemRef(ItemOID="IT.TA.TAETORD", Mandatory="Yes", OrderNumber=5, KeySequence=3),
            DEFINE.ItemRef(ItemOID="IT.TA.ETCD", Mandatory="Yes", OrderNumber=6),
            DEFINE.ItemRef(ItemOID="IT.TA.ELEMENT", Mandatory="No", OrderNumber=7),
            DEFINE.ItemRef(ItemOID="IT.TA.TABRANCH", Mandatory="No", OrderNumber=8),
            DEFINE.ItemRef(ItemOID="IT.TA.TATRANS", Mandatory="No", OrderNumber=9),
            DEFINE.ItemRef(ItemOID="IT.TA.EPOCH", Mandatory="No", OrderNumber=10)
        ]
        return itemrefs

    def create_odm_document(self, igd):
        """
        assemble the ODM document, add the ItemGroupDef, and write it to a file
        :param igd: ItemGroupDef object
        """
        odm = self.create_root()
        study = self.create_study()
        odm.Study = [study]
        mdv = self.create_mdv()
        odm.Study[0].MetaDataVersion = [mdv]
        odm.Study[0].MetaDataVersion[0].ItemGroupDef.append(igd)
        odm.write_xml(self.test_file)
        return odm

    def create_odm_document_json(self, igd):
        """
        assemble the ODM document, add the ItemGroupDef, and write it to a file
        :param igd: ItemGroupDef object
        """
        odm = self.create_root()
        study = self.create_study()
        odm.Study = [study]
        mdv = self.create_mdv()
        odm.Study[0].MetaDataVersion = [mdv]
        odm.Study[0].MetaDataVersion[0].ItemGroupDef.append(igd)
        odm.write_json(self.test_file_json)
        return odm

    def create_root(self):
        """
        create the ODM root element object with test data
        :return: ODM root element object
        """
        root = {"FileOID": "DEFINE.TEST.IGD.001", "FileType": "Snapshot", "AsOfDateTime": self.set_datetime(),
                "CreationDateTime": self.set_datetime(), "ODMVersion": "1.3.2", "Originator": "CDISC 360", "SourceSystem": "WS2",
                "SourceSystemVersion": "0.1", "Context": "Other"}
        root = DEFINE.ODM(**root)
        return root


    @staticmethod
    def create_study():
        """
        create the ODM Study object instantiated with test data
        :return: ODM Study element object
        """
        study = DEFINE.Study(OID="ST.TEST.IGD.001")
        study.GlobalVariables.StudyName = DEFINE.StudyName(_content="TEST ODM ItemGroupDef")
        study.GlobalVariables.StudyDescription = DEFINE.StudyDescription(_content="ItemGroupDef 001")
        study.GlobalVariables.ProtocolName = DEFINE.ProtocolName(_content="ODM ItemGroupDef")
        return study


    @staticmethod
    def create_mdv():
        """
        create the ODM MetaDataVersion object instantiated with test data
        :return: ODM MetaDataVersion element object
        """
        mdv = DEFINE.MetaDataVersion(OID="MDV.TEST.IGD.001", Name="ItemGroupDefTest001",
                                     Description="ItemGroupDef Test 001", DefineVersion="2.1.0")
        return mdv


    @staticmethod
    def write_odm_file(odm, odm_file):
        """
        write the ODM document to a file
        :param odm: ODM document
        :param odm_file: path and name of ODM file to write ODM document to
        """
        tree = ET.ElementTree(odm)
        tree.write(odm_file, xml_declaration=True)


    @staticmethod
    def set_datetime():
        """return the current datetime in ISO 8601 format"""
        return datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()



from abc import ABC, abstractmethod
from cerberus import schema_registry, validator


class ConformanceChecker(ABC):
    @abstractmethod
    def verify_conformance(self, doc, schema_name):
        raise NotImplementedError(
            "Attempted to execute an abstract method validate_tree in the Validator class")


class MetadataSchema(ConformanceChecker):
    def __init__(self):
        self._set_metadata_registry()

    def verify_conformance(self, doc, schema_name):
        schema = schema_registry.get(schema_name)
        v = validator.Validator(schema)
        is_valid = v.validate(doc)
        if not is_valid:
            raise ValueError(v.errors)
        return is_valid

    @staticmethod
    def _set_metadata_registry():
        """ a cerberus json schema has been generated from the odm_1_3_2 model """
        schema_registry.add("TranslatedText", {"lang": {"type": "string"},
                                                 "_content": {"type": "string", "required": True}})

        schema_registry.add("Alias", {
            "Context": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True}
        })

        schema_registry.add("Description", {"TranslatedText": {"type": "list",
                            "schema": {"type": "dict", "schema": schema_registry.get("TranslatedText")}}})

        schema_registry.add("StudyEventRef", {
                    "StudyEventOID": {"type": "string", "required": True},
                    "OrderNumber":  {"type": "integer", "required": False},
                    "Mandatory": {"type": "string", "allowed": ["Yes", "No"]},
                    "CollectionExceptionOID": {"type": "string", "required": False}
                })

        schema_registry.add("Protocol", {
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "StudyEventRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("StudyEventRef")}},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("FormRef", {
            "FormOID": {"type": "string", "required": True},
            "OrderNumber":  {"type": "integer", "required": False},
            "Mandatory": {"type": "string", "allowed": ["Yes", "No"]},
            "CollectionExceptionOID": {"type": "string", "required": False}
        })

        schema_registry.add("StudyEventDef", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "Repeating": {"type": "string", "allowed": ["Yes", "No"]},
            "Type": {"type": "string", "allowed": ["Scheduled", "Unscheduled", "Common"]},
            "Category": {"type": "string"},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "FormRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("FormRef")}},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("ItemGroupRef", {
            "ItemGroupOID": {"type": "string", "required": True},
            "OrderNumber": {"type": "integer", "required": False},
            "Mandatory": {"type": "string", "allowed": ["Yes", "No"]},
            "CollectionExceptionOID": {"type": "string", "required": False}
        })

        schema_registry.add("ArchiveLayout", {
            "OID": {"type": "string", "required": True},
            "PdfFileName": {"type": "string", "required": True},
        })

        schema_registry.add("FormDef", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "Repeating": {"type": "string", "allowed": ["Yes", "No"]},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "ItemGroupRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("ItemGroupRef")}},
            "ArchiveLayout": {"type": "list","schema": {"type": "dict", "schema": schema_registry.get("ArchiveLayout")}},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("ItemRef", {
            "ItemOID": {"type": "string", "required": True},
            "OrderNumber":  { "type": "integer", "required": False},
            "Mandatory": { "type": "string", "required": False, "allowed": ["Yes", "No"]},
            "KeySequence": { "type": "integer", "required": False},
            "MethodOID": {"type": "string", "required": False},
            "Role": {"type": "string", "required": False},
            "RoleCodeListOID": {"type": "string", "required": False},
            "CollectionExceptionOID": {"type": "string", "required": False},
        })

        schema_registry.add("ItemGroupDef", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "Repeating": {"type": "string", "allowed": ["Yes", "No"]},
            "IsReferenceData": {"type": "string", "allowed": ["Yes", "No"]},
            "SASDatasetName": {"type": "string"},
            "Domain": {"type": "string"},
            "Origin": {"type": "string"},
            "Purpose": {"type": "string"},
            "Comment": {"type": "string"},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "ItemRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("ItemRef")}},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("FormalExpression", {
            "Context": {"type": "string", "required": True},
            "_content": {"type": "string", "required": True}
        })

        schema_registry.add("MeasurementUnitRef", {
            "MeasurementUnitOID": {"type": "string", "required": True}
        })

        schema_registry.add("RangeCheck", {
            "Comparator": {"type": "string", "allowed": ["LT", "LE", "GT", "GE", "EQ", "NE", "IN", "NOTIN"]},
            "SoftHard": {"type": "string", "allowed": ["Soft", "Hard"]},
            "CheckValue": {"type": "list", "schema": {"type": "dict", "schema": {"_content": {"type": "string"}}}},
            "FormalExpression": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("FormalExpression")}},
            "MeasurementUnitRef": {"type": "dict", "schema": schema_registry.get("MeasurementUnitRef")},
            "ErrorMessage": {"type": "dict", "schema": {"TranslatedText": {"type": "list",
                         "schema": {"type": "dict", "schema": schema_registry.get("TranslatedText")}}}}
        })

        schema_registry.add("ExternalQuestion", {
            "Dictionary": {"type": "string", "required": False},
            "Version": {"type": "string", "required": False},
            "Code": {"type": "string", "required": False},
        })

        schema_registry.add("CodeListRef", {
            "CodeListOID": {"type": "string"}
        })

        schema_registry.add("ItemDef", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "DataType": {"type": "string", "allowed": ["text", "integer", "float", "date", "time", "datetime", "string",
                                                        "boolean", "double", "hexBinary", "base64Binary", "hexFloat",
                                                        "base64Float", "partialDate", "partialTime", "partialDatetime",
                                                        "durationDatetime", "intervalDatetime", "incompleteDatetime",
                                                        "incompleteDate", "incompleteTime", "URI"]},
            "Length": {"type": "integer"},
            "SignificantDigits": {"type": "integer"},
            "SASFieldName": {"type": "string"},
            "SDSVarName": {"type": "string"},
            "Origin": {"type": "string"},
            "Comment": {"type": "string"},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "Question": {"type": "dict", "schema": {"TranslatedText": {"type": "list",
                         "schema": {"type": "dict", "schema": schema_registry.get("TranslatedText")}}}},
            "ExternalQuestion": {"type": "dict", "schema": schema_registry.get("ExternalQuestion")},
            "CodeListRef": {"type": "dict", "schema": schema_registry.get("CodeListRef")},
            "ErrorMessage": {"type": "dict", "schema": {"TranslatedText": {"type": "list",
                             "schema": {"type": "dict", "schema": schema_registry.get("TranslatedText")}}}},
            "RangeCheck": {"type": "dict", "schema": schema_registry.get("RangeCheck")},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("CodeListItem", {
            "CodedValue": {"type": "string", "required": True},
            "Rank": {"type": "float"},
            "OrderNumber": {"type": "integer"},
            "Decode": {"type": "dict", "schema": {"TranslatedText": {"type": "list",
                       "schema": {"type": "dict", "schema": schema_registry.get("TranslatedText")}}}},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("EnumeratedItem", {
            "CodedValue": {"type": "string", "required": True},
            "Rank": {"type": "float"},
            "OrderNumber": {"type": "integer"},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("ExternalCodeList", {
            "Dictionary": {"type": "string"},
            "Version": {"type": "string"},
            "ref": {"type": "string"},
            "href": {"type": "string"}
        })

        schema_registry.add("CodeList", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "DataType": {"type": "string", "allowed": ["text", "integer", "float", "string"]},
            "SASFormatName": {"type": "string"},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "CodeListItem": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("CodeListItem")}},
            "EnumeratedItem": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("EnumeratedItem")}},
            "ExternalCodeList": {"type": "dict", "schema": schema_registry.get("ExternalCodeList")},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("ConditionDef", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "FormalExpression": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("FormalExpression")}},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("MethodDef", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "Type": {"type": "string", "required": True, "allowed": ["Computation", "Imputation", "Transpose", "Other"]},
            "Description": {"type": "dict", "required": True, "schema": schema_registry.get("Description")},
            "FormalExpression": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("FormalExpression")}},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("MetaDataVersion", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "Description": {"type": "string"},
            "Protocol": {"type": "dict", "schema": schema_registry.get("Protocol")},
            "StudyEventDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("StudyEventDef")}},
            "FormDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("FormDef")}},
            "ItemGroupDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("ItemGroupDef")}},
            "ItemDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("ItemDef")}},
            "CodeList": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("CodeList")}},
            "ConditionDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("ConditionDef")}},
            "MethodDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("MethodDef")}}
        })


        schema_registry.add("Study", {
            "OID": {"type": "string", "required": True},
            "GlobalVariables": {"type": "dict", "required": True, "schema": {
                    "StudyName": {"schema": {"_content": {"type": "string", "required": True}}},
                    "StudyDescription": {"schema": {"_content": {"type": "string", "required": True}}},
                    "ProtocolName": {"schema": {"_content": {"type": "string", "required": True}}}
                }
            },
            "MetaDataVersion": {"type": "dict", "schema": schema_registry.get("MetaDataVersion")}
        })

        # schema_registry.add("Study", {
        #     "OID": {"type": "string", "required": True},
        #     "GlobalVariables": {"type": "dict", "required": True, "schema": {
        #             "StudyName": {"schema": {"TranslatedText": {"type": "list",
        #                           "schema": {"type": "dict", "schema": schema_registry.get("TranslatedText")}}}},
        #             "StudyDescription": {"schema": {"TranslatedText": {"type": "list",
        #                                  "schema": {"type": "dict", "schema": schema_registry.get("TranslatedText")}}}},
        #             "ProtocolName": {"schema": {"TranslatedText": {"type": "list",
        #                              "schema": {"type": "dict", "schema": schema_registry.get("TranslatedText")}}}}
        #         }
        #     },
        #     "MetaDataVersion": {"type": "dict", "schema": schema_registry.get("MetaDataVersion")}
        # })



import unittest
import odmlib.odm_1_3_2.model as ODM
import odmlib.odm_1_3_2.rules.metadata_schema as METADATA


class TestAttributeOrder(unittest.TestCase):
    def setUp(self):
        self.validator = METADATA.MetadataSchema()

    def test_study_invalid_study_object(self):
        study = ODM.Study(OID="ST.001.Test")
        # StudyName is a child of GLobalVariables, not of Study
        with self.assertRaises(TypeError):
            study.StudyName = ODM.StudyName(_content="The ODM study name")

    def test_study_invalid_study_object_assignment(self):
        study = ODM.Study(OID="ST.001.Test")
        study.GlobalVariables = ODM.GlobalVariables()
        with self.assertRaises(TypeError):
            study.GlobalVariables.StudyName = ODM.ProtocolName(_content="The ODM protocol name")

    def test_valid_order_study_objects(self):
        study = ODM.Study(OID="ST.001.Test")
        study.GlobalVariables = ODM.GlobalVariables()
        study.GlobalVariables.StudyName = ODM.StudyName(_content="The ODM study name")
        study.GlobalVariables.StudyDescription = ODM.StudyDescription(_content="The description of the ODM study")
        study.GlobalVariables.ProtocolName = ODM.ProtocolName(_content="The ODM protocol name")
        self.assertTrue(study.verify_order())

    def test_invalid_order_study_objects(self):
        study = ODM.Study(OID="ST.001.Test")
        study.GlobalVariables = ODM.GlobalVariables()
        study.GlobalVariables.ProtocolName = ODM.ProtocolName(_content="The ODM protocol name")
        study.GlobalVariables.StudyName = ODM.StudyName(_content="The ODM study name")
        study.GlobalVariables.StudyDescription = ODM.StudyDescription(_content="The description of the ODM study")
        with self.assertRaises(ValueError):
            study.verify_order()

    def test_odm_order_incomplete(self):
        attrs = self.get_root_attributes()
        odm = ODM.ODM(**attrs)
        odm.Study.append(ODM.Study(OID="ST.001.Test"))
        odm.Study[0].GlobalVariables = ODM.GlobalVariables()
        odm.Study[0].GlobalVariables.StudyName = ODM.StudyName(_content="The ODM study name")
        odm.Study[0].GlobalVariables.StudyDescription = ODM.StudyDescription(_content="The description of the ODM study")
        odm.Study[0].GlobalVariables.ProtocolName = ODM.ProtocolName(_content="The ODM protocol name")
        attrs = self.get_mdv_attributes()
        odm.Study[0].MetaDataVersion.append(ODM.MetaDataVersion(**attrs))
        odm.Study[0].verify_order()

    def test_odm_order_multiple(self):
        attrs = self.get_root_attributes()
        odm = ODM.ODM(**attrs)
        odm.Study.append(ODM.Study(OID="ST.001.Test"))
        odm.Study[0].GlobalVariables = ODM.GlobalVariables()
        odm.Study[0].GlobalVariables.StudyName = ODM.StudyName(_content="The ODM study name")
        odm.Study[0].GlobalVariables.StudyDescription = ODM.StudyDescription(_content="The description of the ODM study")
        odm.Study[0].GlobalVariables.ProtocolName = ODM.ProtocolName(_content="The ODM protocol name")
        attrs = self.get_mdv_attributes()
        odm.Study[0].MetaDataVersion.append(ODM.MetaDataVersion(**attrs))
        fd1 = ODM.FormDef(OID="ODM.F.VS", Name="Vital Signs Form", Repeating="Yes")
        fd1.Description = ODM.Description()
        fd1.Description.TranslatedText.append(ODM.TranslatedText(lang="en", _content="this is the test description"))
        fd1.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.COMMON", Mandatory="Yes", OrderNumber=1))
        fd1.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS_GENERAL", Mandatory="Yes", OrderNumber=2))
        fd1.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS", Mandatory="Yes", OrderNumber=3))
        fd2 = ODM.FormDef(OID="ODM.F.DM", Name="Demographics Form", Repeating="No")
        fd3 = ODM.FormDef(OID="ODM.F.AE", Name="Adverse Events Form", Repeating="Yes")
        odm.Study[0].MetaDataVersion[0].FormDef = [fd1, fd2, fd3]
        odm.Study[0].verify_order()

    def test_odm_order_itemdef(self):
        # create an ItemDef
        itd = ODM.ItemDef(OID="ODM.IT.DM.BRTHYR", Name="Birth Year", DataType="integer")
        itd.Description = ODM.Description()
        itd.Description.TranslatedText.append(ODM.TranslatedText(_content="Year of the subject's birth", lang="en"))
        itd.Question = ODM.Question()
        itd.Question.TranslatedText.append(ODM.TranslatedText(_content="Birth Year", lang="en"))
        itd.Alias.append(ODM.Alias(Context="CDASH", Name="BRTHYR"))
        itd.Alias.append(ODM.Alias(Context="SDTM", Name="BRTHDTC"))
        self.assertTrue(itd.verify_order())

    def test_odm_order_itemdef_invalid(self):
        # create an ItemDef
        attrs = self.get_root_attributes()
        odm = ODM.ODM(**attrs)
        odm.Study.append(ODM.Study(OID="ST.001.Test"))
        odm.Study[0].GlobalVariables = ODM.GlobalVariables()
        odm.Study[0].GlobalVariables.StudyName = ODM.StudyName(_content="The ODM study name")
        odm.Study[0].GlobalVariables.StudyDescription = ODM.StudyDescription(_content="The description of the ODM study")
        odm.Study[0].GlobalVariables.ProtocolName = ODM.ProtocolName(_content="The ODM protocol name")
        attrs = self.get_mdv_attributes()
        odm.Study[0].MetaDataVersion.append(ODM.MetaDataVersion(**attrs))
        itd = ODM.ItemDef(OID="ODM.IT.DM.BRTHYR", Name="Birth Year", DataType="integer")
        itd.Alias.append(ODM.Alias(Context="CDASH", Name="BRTHYR"))
        itd.Alias.append(ODM.Alias(Context="SDTM", Name="BRTHDTC"))
        itd.Description = ODM.Description()
        itd.Description.TranslatedText.append(ODM.TranslatedText(_content="Year of the subject's birth", lang="en"))
        itd.Question = ODM.Question()
        itd.Question.TranslatedText.append(ODM.TranslatedText(_content="Birth Year", lang="en"))
        odm.Study[0].MetaDataVersion[0].ItemDef.append(itd)
        with self.assertRaises(ValueError):
            self.assertTrue(odm.Study[0].MetaDataVersion[0].verify_order())

    def test_reordering_object(self):
        itd = ODM.ItemDef(OID="ODM.IT.DM.BRTHYR", Name="Birth Year", DataType="integer")
        itd.Alias.append(ODM.Alias(Context="CDASH", Name="BRTHYR"))
        itd.Alias.append(ODM.Alias(Context="SDTM", Name="BRTHDTC"))
        itd.Description = ODM.Description()
        itd.Description.TranslatedText.append(ODM.TranslatedText(_content="Year of the subject's birth", lang="en"))
        itd.Question = ODM.Question()
        itd.Question.TranslatedText.append(ODM.TranslatedText(_content="Birth Year", lang="en"))
        with self.assertRaises(ValueError):
            itd.verify_order()
        itd.reorder_object()
        self.assertTrue(itd.verify_order())

    def test_reorder_study_objects(self):
        study = ODM.Study(OID="ST.001.Test")
        study.GlobalVariables = ODM.GlobalVariables()
        study.GlobalVariables.ProtocolName = ODM.ProtocolName(_content="The ODM protocol name")
        study.GlobalVariables.StudyName = ODM.StudyName(_content="The ODM study name")
        study.GlobalVariables.StudyDescription = ODM.StudyDescription(_content="The description of the ODM study")
        with self.assertRaises(ValueError):
            study.verify_order()
        study.GlobalVariables.reorder_object()
        self.assertTrue(study.verify_order())

    def test_reorder_mdv_object(self):
        attrs = self.get_mdv_attributes()
        mdv = ODM.MetaDataVersion(**attrs)
        mdv.StudyEventDef = self.add_SED()
        mdv.FormDef = self.add_FD()
        mdv.ItemGroupDef = self.add_IGD()
        mdv.ItemDef = self.add_ITD()
        self.assertTrue(mdv.verify_order())
        mdv2 = ODM.MetaDataVersion(**attrs)
        mdv2.StudyEventDef = self.add_SED()
        mdv2.ItemDef = self.add_ITD()
        mdv2.ItemGroupDef = self.add_IGD()
        mdv2.FormDef = self.add_FD()
        with self.assertRaises(ValueError):
            mdv2.verify_order()
        mdv2.reorder_object()
        self.assertTrue(mdv2.verify_order())

    def test_reorder_mdv_protocol_objects(self):
        attrs = self.get_mdv_attributes()
        mdv = ODM.MetaDataVersion(**attrs)
        mdv.Protocol = self.add_protocol()
        mdv.MethodDef = self.add_MD()
        mdv.StudyEventDef = self.add_SED()
        mdv.FormDef = self.add_FD()
        mdv.CodeList = self.add_CL()
        mdv.ItemGroupDef = self.add_IGD()
        mdv.ItemDef = self.add_ITD()
        # test mdv order
        with self.assertRaises(ValueError):
            mdv.verify_order()
        mdv.reorder_object()
        # test protocol order
        with self.assertRaises(ValueError):
            mdv.verify_order()
        mdv.Protocol.reorder_object()
        self.assertTrue(mdv.verify_order())

    def get_mdv_attributes(self):
        return {"OID": "MDV.TRACE-XML-ODM-01", "Name": "TRACE-XML MDV", "Description": "Trace-XML Example"}

    def get_root_attributes(self):
        return {"FileOID": "ODM.MDV.TEST.001", "Granularity": "Metadata",
                 "AsOfDateTime": "2020-07-13T00:13:51.309617+00:00",
                 "CreationDateTime": "2020-07-13T00:13:51.309617+00:00", "ODMVersion": "1.3.2", "FileType": "Snapshot",
                 "Originator": "RDS", "SourceSystem": "ODMLib", "SourceSystemVersion": "0.1",
                 "schemaLocation": "http://www.cdisc.org/ns/odm/v1.3 odm1-3-2.xsd"}

    def add_ITD(self):
        # ItemDef 1
        ttd1 = ODM.TranslatedText(_content="Date of measurements", lang="en")
        ttq1 = ODM.TranslatedText(_content="Date", lang="en")
        desc1 = ODM.Description()
        desc1.TranslatedText = [ttd1]
        q1 = ODM.Question()
        q1.TranslatedText = [ttq1]
        a1 = ODM.Alias(Context="CDASH", Name="VSDAT")
        itd1 = ODM.ItemDef(OID="ODM.IT.VS.VSDAT", Name="Date", DataType="partialDate")
        itd1.Description = desc1
        itd1.Question = q1
        itd1.Alias = [a1]
        # ItemDef 2
        ttd2 = ODM.TranslatedText(_content="Result of the vital signs measurement as originally received or collected.", lang="en")
        ttq2 = ODM.TranslatedText(_content="Diastolic", lang="en")
        desc2 = ODM.Description()
        desc2.TranslatedText = [ttd2]
        q2 = ODM.Question()
        q2.TranslatedText = [ttq2]
        a2a = ODM.Alias(Context="CDASH", Name="BP.DIABP.VSORRES")
        a2b = ODM.Alias(Context="CDASH/SDTM", Name="VSORRES+VSORRESU")
        itd2 = ODM.ItemDef(OID="ODM.IT.VS.BP.VSORRESU", Name="BP Units", DataType="text")
        itd2.Description = desc2
        itd2.Question = q2
        itd2.Alias = [a2a, a2b]
        return [itd1, itd2]

    def add_IGD(self):
        itr1 = ODM.ItemRef(ItemOID="ODM.IT.VS.VSDAT", Mandatory="Yes")
        itr2 = ODM.ItemRef(ItemOID="ODM.IT.VS.BP.DIABP.VSORRES", Mandatory="Yes")
        itr3 = ODM.ItemRef(ItemOID="ODM.IT.VS.BP.SYSBP.VSORRES", Mandatory="Yes")
        igd1 = ODM.ItemGroupDef(OID="ODM.IG.VS", Name="Vital Sign Measurement", Repeating="Yes")
        igd1.ItemRef = [itr1, itr2, itr3]
        igr4 = ODM.ItemRef(ItemOID="ODM.IT.DM.BRTHYR", Mandatory="Yes")
        igr5 = ODM.ItemRef(ItemOID="ODM.IT.DM.SEX", Mandatory="Yes")
        igd2 = ODM.ItemGroupDef(OID="ODM.IG.DM", Name="Demographics", Repeating="No")
        igd2.ItemRef = [igr4, igr5]
        return [igd1, igd2]

    def add_FD(self):
        igr1 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.Common", Mandatory="Yes", OrderNumber=1)
        igr2 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS_GENERAL", Mandatory="Yes", OrderNumber=2)
        igr3 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS", Mandatory="Yes", OrderNumber=3)
        fd1 = ODM.FormDef(OID="ODM.F.VS", Name="Vital Signs", Repeating="No")
        fd1.ItemGroupRef = [igr1, igr2, igr3]
        igr4 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.Common", Mandatory="Yes", OrderNumber=1)
        igr5 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.DM", Mandatory="Yes", OrderNumber=2)
        fd2 = ODM.FormDef(OID="ODM.F.DM", Name="Demographics", Repeating="No")
        fd2.ItemGroupRef = [igr4, igr5]
        return [fd1, fd2]

    def add_SED(self):
        fr1 = ODM.FormRef(FormOID="ODM.F.DM", Mandatory="Yes", OrderNumber=1)
        fr2 = ODM.FormRef(FormOID="ODM.F.VS", Mandatory="Yes", OrderNumber=2)
        fr3 = ODM.FormRef(FormOID="ODM.F.AE", Mandatory="Yes", OrderNumber=3)
        ser1 = ODM.StudyEventDef(OID="BASELINE", Name="Baseline Visit", Repeating="No", Type="Scheduled")
        ser1.FormRef = [fr1, fr2, fr3]
        ser2 = ODM.StudyEventDef(OID="FOLLOW-UP", Name="Follow-up Visit", Repeating="Yes", Type="Scheduled")
        ser2.FormRef = [fr1, fr2, fr3]
        return [ser1, ser2]

    def add_MD(self):
        tt1 = ODM.TranslatedText(_content="Concatenation of BRTHYR, BRTHMO, and BRTHDY in ISO 8601 format", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt1]
        md = ODM.MethodDef(OID="ODM.MT.DOB", Name="Create BRTHDTC from date ELEMENTS", Type="Computation")
        md.Description = desc
        return [md]

    def add_CL(self):
        tt1 = ODM.TranslatedText(_content="No", lang="en")
        dc1 = ODM.Decode()
        dc1.TranslatedText = [tt1]
        cli1 = ODM.CodeListItem(CodedValue="N")
        cli1.Decode = dc1
        tt2 = ODM.TranslatedText(_content="Yes", lang="en")
        dc2 = ODM.Decode()
        dc2.TranslatedText = [tt2]
        cli2 = ODM.CodeListItem(CodedValue="Y")
        cli2.Decode = dc2
        cl = ODM.CodeList(OID="ODM.CL.NY_SUB_Y_N", Name="No Yes Response", DataType="text")
        cl.CodeListItem = [cli1, cli2]
        return [cl]

    def add_protocol(self):
        p = ODM.Protocol()
        # invalid order for protocol content to test re-order
        p.Alias = [ODM.Alias(Context="ClinicalTrials.gov", Name="trace-protocol")]
        tt = ODM.TranslatedText(_content="Trace-XML Test CDASH File", lang="en")
        p.Description = ODM.Description()
        p.Description.TranslatedText = [tt]
        ser1 = ODM.StudyEventRef(StudyEventOID="BASELINE", OrderNumber=1, Mandatory="Yes")
        ser2 = ODM.StudyEventRef(StudyEventOID="FOLLOW-UP", OrderNumber=2, Mandatory="Yes")
        p.StudyEventRef = [ser1, ser2]
        return p



import validators


class Borg:
    """ borg provides a set of global attributes for the Namespace Registry singleton """
    namespaces = {}
    default_namespace = {}

    @classmethod
    def reset(cls):
        cls.namespaces = {}
        cls.default_namespace = {}


class NamespaceRegistry(Borg):
    def __init__(self, prefix=None, uri=None, is_default=False, is_reset=False):
        if is_reset:
            super().reset()
        if prefix is not None and uri is not None:
            if validators.url(uri):
                self._update_registry(prefix, uri, is_default)
            else:
                raise ValueError(f"Namespace uri is not a valid url: {uri}")

    def get_odm_namespace_entries(self):
        entries = ["xmlns=" + list(self.default_namespace.values())[0]]
        for prefix, uri in self.namespaces.items():
            if prefix != list(self.default_namespace.keys())[0]:
                entries.append("xmlns:" + prefix + "=" + uri)
        return entries

    def get_ns_entry_dict(self, prefix):
        if prefix in self.namespaces:
            return {prefix: self.namespaces[prefix]}
        else:
            return {}

    def get_ns_attribute_name(self, name, prefix):
        if prefix in self.namespaces:
            if prefix in self.default_namespace:
                return name
            else:
                return "{" + self.namespaces[prefix] + "}" + name
        else:
            raise ValueError(f"Error: Namespace with prefix {prefix} has not been registered")

    def get_prefix_ns_from_uri(self, uri):
        for prefix, ns_uri in self.namespaces.items():
            if uri.lower() == ns_uri.lower():
                return prefix
        raise ValueError(f"Error: Namespace with URI {uri} has not been registered")

    def set_odm_namespace_attributes(self, odm_elem):
        """ add NS attributes to the ODM XML element object """
        odm_elem.attrib["xmlns"] = list(self.default_namespace.values())[0]
        for prefix, uri in self.namespaces.items():
            if prefix != list(self.default_namespace.keys())[0]:
                odm_elem.attrib["xmlns:" + prefix] = uri

    def _update_registry(self, prefix, uri, is_default):
        self.namespaces[prefix] = uri
        if is_default:
            self.default_namespace[prefix] = uri

    def remove_registry_entry(self, prefix):
        if prefix in self.namespaces:
            self.namespaces.pop(prefix)
        if prefix in self.default_namespace:
            self.default_namespace.pop(prefix)



from abc import ABC, abstractmethod
from cerberus import schema_registry, validator


class ConformanceChecker(ABC):
    @abstractmethod
    def check_conformance(self, doc, schema_name):
        raise NotImplementedError(
            "Attempted to execute an abstract method validate_tree in the Validator class")


class MetadataSchema(ConformanceChecker):
    """ The metadata schema for Define-XML v2.0 to aid in conformance checking """
    def __init__(self):
        self._set_metadata_registry()

    def check_conformance(self, doc, schema_name):
        schema = schema_registry.get(schema_name)
        v = validator.Validator(schema)
        is_valid = v.validate(doc)
        if not is_valid:
            raise ValueError(v.errors)
        return is_valid

    @staticmethod
    def _set_metadata_registry():
        schema_registry.add("TranslatedText", {"lang": {"type": "string"},
                                                 "_content": {"type": "string", "required": True}})

        schema_registry.add("Alias", {
            "Context": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True}
        })

        schema_registry.add("Description", {"TranslatedText": {"type": "list",
                            "schema": {"type": "dict", "schema": schema_registry.get("TranslatedText")}}})

        schema_registry.add("title", {"_content": {"type": "string", "required": True}})

        schema_registry.add("leaf", {
            "ID": {"type": "string", "required": True},
            "href": {"type": "string", "required": True},
            "title": {"type": "dict", "schema": schema_registry.get("title")}})

        schema_registry.add("WhereClauseRef", {
            "WhereClauseOID": {"type": "string", "required": True}
        })

        schema_registry.add("ValueListRef", {
            "ValueListOID": {"type": "string", "required": True}
        })

        schema_registry.add("PDFPageRef", {
            "Type": {"type": "string", "required": True, "allowed": ["PhysicalRef", "NamedDestination"]},
            "PageRefs": {"type": "string"},
            "FirstPage": {"type": "integer"},
            "LastPage": {"type": "integer"}
        })

        schema_registry.add("DocumentRef", {
            "leafID": {"type": "string", "required": True},
            "PDFPageRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("PDFPageRef")}}
        })

        schema_registry.add("ItemRef", {
            "ItemOID": {"type": "string", "required": True},
            "OrderNumber":  { "type": "integer", "required": False},
            "Mandatory": { "type": "string", "required": False, "allowed": ["Yes", "No"]},
            "KeySequence": { "type": "integer", "required": False},
            "MethodOID": {"type": "string", "required": False},
            "Role": {"type": "string", "required": False},
            "RoleCodeListOID": {"type": "string", "required": False},
            "WhereClauseRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("WhereClauseRef")}}
        })

        schema_registry.add("ItemGroupDef", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "Repeating": {"type": "string", "allowed": ["Yes", "No"]},
            "IsReferenceData": {"type": "string", "allowed": ["Yes", "No"]},
            "SASDatasetName": {"type": "string"},
            "Domain": {"type": "string"},
            "Origin": {"type": "string"},
            "Purpose": {"type": "string"},
            "Structure": {"type": "string", "required": True},
            "Class": {"type": "string", "required": True},
            "ArchiveLocationID": {"type": "string", "required": True},
            "CommentOID": {"type": "string"},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "ItemRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("ItemRef")}},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}},
            "leaf": {"type": "dict", "schema": schema_registry.get("leaf")}
        })

        schema_registry.add("FormalExpression", {
            "Context": {"type": "string", "required": True},
            "_content": {"type": "string", "required": True}
        })

        schema_registry.add("RangeCheck", {
            "Comparator": {"type": "string", "allowed": ["LT", "LE", "GT", "GE", "EQ", "NE", "IN", "NOTIN"]},
            "SoftHard": {"type": "string", "allowed": ["Soft", "Hard"]},
            "ItemOID": {"type": "string", "required": True},
            "CheckValue": {"type": "list", "schema": {"type": "dict", "schema": {"_content": {"type": "string"}}}}
        })

        schema_registry.add("Origin", {
            "Type": {"type": "string", "required": True,
                     "allowed": ["CRF", "Derived", "Assigned", "Assigned", "Protocol", "eDT", "Predecessor"]},
            "DocumentRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("DocumentRef")}},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")}
        })

        schema_registry.add("CodeListRef", {
            "CodeListOID": {"type": "string"}
        })

        schema_registry.add("ItemDef", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "DataType": {"type": "string", "allowed": ["text", "integer", "float", "date", "time", "datetime", "string",
                                                        "boolean", "double", "hexBinary", "base64Binary", "hexFloat",
                                                        "base64Float", "partialDate", "partialTime", "partialDatetime",
                                                        "durationDatetime", "intervalDatetime", "incompleteDatetime",
                                                        "incompleteDate", "incompleteTime", "URI"]},
            "Length": {"type": "integer"},
            "SignificantDigits": {"type": "integer"},
            "SASFieldName": {"type": "string"},
            "DisplayFormat": {"type": "string"},
            "CommentOID": {"type": "string"},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "CodeListRef": {"type": "dict", "schema": schema_registry.get("CodeListRef")},
            "Origin": {"type": "dict", "schema": schema_registry.get("Origin")},
            "ValueListRef": {"type": "dict", "schema": schema_registry.get("ValueListRef")},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("CodeListItem", {
            "CodedValue": {"type": "string", "required": True},
            "Rank": {"type": "float"},
            "OrderNumber": {"type": "integer"},
            "ExtendedValue": {"type": "string", "allowed": ["Yes"]},
            "Decode": {"type": "dict", "schema": {"TranslatedText": {"type": "list",
                       "schema": {"type": "dict", "schema": schema_registry.get("TranslatedText")}}}},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("EnumeratedItem", {
            "CodedValue": {"type": "string", "required": True},
            "Rank": {"type": "float"},
            "OrderNumber": {"type": "integer"},
            "ExtendedValue": {"type": "string", "allowed": ["Yes"]},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("ExternalCodeList", {
            "Dictionary": {"type": "string"},
            "Version": {"type": "string"},
            "ref": {"type": "string"},
            "href": {"type": "string"}
        })

        schema_registry.add("CodeList", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "DataType": {"type": "string", "allowed": ["text", "integer", "float", "string"]},
            "SASFormatName": {"type": "string"},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "CodeListItem": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("CodeListItem")}},
            "EnumeratedItem": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("EnumeratedItem")}},
            "ExternalCodeList": {"type": "dict", "schema": schema_registry.get("ExternalCodeList")},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}}
        })

        schema_registry.add("AnnotatedCRF", {
            "DocumentRef": {"type": "dict", "schema": schema_registry.get("DocumentRef")}
        })

        schema_registry.add("SupplementalDoc", {
            "DocumentRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("DocumentRef")}}
        })

        schema_registry.add("WhereClauseDef", {
            "OID": {"type": "string", "required": True},
            "CommentOID": {"type": "string"},
            "RangeCheck": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("RangeCheck")}}
        })

        schema_registry.add("ValueListDef", {
            "OID": {"type": "string", "required": True},
            "ItemRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("ItemRef")}}
        })

        schema_registry.add("CommentDef", {
            "OID": {"type": "string", "required": True},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "DocumentRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("DocumentRef")}}
        })

        schema_registry.add("MethodDef", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "Type": {"type": "string", "required": True, "allowed": ["Computation", "Imputation", "Transpose", "Other"]},
            "Description": {"type": "dict", "schema": schema_registry.get("Description")},
            "FormalExpression": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("FormalExpression")}},
            "Alias": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("Alias")}},
            "DocumentRef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("DocumentRef")}}
        })

        schema_registry.add("MetaDataVersion", {
            "OID": {"type": "string", "required": True},
            "Name": {"type": "string", "required": True},
            "Description": {"type": "string"},
            "DefineVersion": {"type": "string", "required": True},
            "StandardName": {"type": "string", "required": True},
            "StandardVersion": {"type": "string", "required": True},
            "AnnotatedCRF": {"type": "dict", "schema": schema_registry.get("AnnotatedCRF")},
            "SupplementalDoc": {"type": "dict", "schema": schema_registry.get("SupplementalDoc")},
            "ValueListDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("ValueListDef")}},
            "WhereClauseDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("WhereClauseDef")}},
            "ItemGroupDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("ItemGroupDef")}},
            "ItemDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("ItemDef")}},
            "CodeList": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("CodeList")}},
            "MethodDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("MethodDef")}},
            "CommentDef": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("CommentDef")}},
            "leaf": {"type": "list", "schema": {"type": "dict", "schema": schema_registry.get("leaf")}}
        })

        schema_registry.add("StudyName", {"_content": {"type": "string", "required": True}})

        schema_registry.add("StudyDescription", {"_content": {"type": "string", "required": True}})

        schema_registry.add("ProtocolName", {"_content": {"type": "string", "required": True}})

        schema_registry.add("GlobalVariables", {
            "StudyName": {"type": "dict", "schema": schema_registry.get("StudyName")},
            "StudyDescription": {"type": "dict", "schema": schema_registry.get("StudyDescription")},
            "ProtocolName": {"type": "dict", "schema": schema_registry.get("ProtocolName")}
        })

        schema_registry.add("Study", {
            "OID": {"type": "string", "required": True},
            "GlobalVariables": {"type": "dict", "schema": schema_registry.get("GlobalVariables")},
            "MetaDataVersion": {"type": "dict", "schema": schema_registry.get("MetaDataVersion")}
        })



from unittest import TestCase
import json
import odmlib.odm_1_3_2.model as ODM


class TestEnumeratedItem(TestCase):
    def setUp(self) -> None:
        attrs = self._set_attributes()
        self.eni = ODM.EnumeratedItem(**attrs)

    def test_add_alias(self):
        self.eni.Alias.append(ODM.Alias(Context="nci:ExtCodeID", Name="C64848"))
        self.assertEqual(self.eni.Alias[0].Context, "nci:ExtCodeID")
        self.assertEqual(self.eni.Alias[0].Name, "C64848")

    def test_to_json(self):
        attrs = self._set_attributes()
        eni = ODM.EnumeratedItem(**attrs)
        eni.Alias.append(ODM.Alias(Context="nci:ExtCodeID", Name="C64848"))
        eni_json = eni.to_json()
        eni_dict = json.loads(eni_json)
        self.assertEqual(eni_dict["CodedValue"], "HGB")
        self.assertDictEqual(eni_dict, self.expected_dict())

    def test_to_dict(self):
        attrs = self._set_attributes()
        eni = ODM.EnumeratedItem(**attrs)
        eni.Alias.append(ODM.Alias(Context="nci:ExtCodeID", Name="C64848"))
        eni_dict = eni.to_dict()
        self.assertEqual(eni_dict["CodedValue"], "HGB")
        self.assertDictEqual(eni_dict, self.expected_dict())

    def test_to_xml(self):
        attrs = self._set_attributes()
        eni = ODM.EnumeratedItem(**attrs)
        eni.Alias.append(ODM.Alias(Context="nci:ExtCodeID", Name="C64848"))
        eni_xml = eni.to_xml()
        self.assertEqual(eni_xml.attrib["CodedValue"], "HGB")
        alias_elem = eni_xml.find("Alias")
        self.assertEqual(alias_elem.attrib["Name"], "C64848")

    @staticmethod
    def _set_attributes():
        """
        set some EnumeratedItem element attributes using test data
        :return: (dict) dictionary with EnumeratedItem attribute information
        """
        return {"CodedValue": "HGB", "OrderNumber": 1}

    @staticmethod
    def expected_dict():
        return {'CodedValue': 'HGB', 'OrderNumber': 1, 'Alias': [{'Context': 'nci:ExtCodeID', 'Name': 'C64848'}]}




class ValueSet:
    _value_set = {
        "Branching.Type": ["Exclusive", "Parallel"],
        "RelativeTimingConstraint.Type": ["StartToStart", "StartToFinish", "FinishToStart", "FinishToFinish"],
        "PDFPageRef.Type": ["NamedDestination", "PhysicalRef"],
        "Origin.Type": ["Collected", "Derived", "Assigned", "Protocol", "Not Available"],
        "Origin.Source": ["Subject", "Investigator", "Vendor", "Sponsor"],
        "RangeCheck.Comparator": ["LT", "LE", "GT", "GE", "EQ", "NE", "IN", "NOTIN"],
        "RangeCheck.SoftHard": ["Soft", "Hard"],
        "StudyEventDef.Repeating": ["Yes", "No"],
        "CodeListItem.ExtendedValue": ["Yes"],
        "EnumeratedItem.ExtendedValue": ["Yes"],
        "FormDef.Repeating": ["Yes", "No"],
        "ItemGroupDef.Repeating": ["Yes", "No"],
        "ItemGroupDef.Type": ["Form", "Dataset", "Concept"],
        "FormRef.Mandatory": ["Yes", "No"],
        "ItemGroupRef.Mandatory": ["Yes", "No"],
        "StudyEventGroupRef.Mandatory": ["Yes", "No"],
        "ItemRef.Mandatory": ["Yes", "No"],
        "StudyEventRef.Mandatory": ["Yes", "No"],
        "ItemGroupDef.IsReferenceData": ["Yes", "No"],
        "ItemDef.DataType": ["text", "integer", "float", "date", "time", "datetime", "string", "boolean", "double",
                             "hexBinary", "base64Binary", "hexFloat", "base64Float", "partialDate", "partialTime",
                             "partialDatetime", "durationDatetime", "intervalDatetime", "incompleteDatetime",
                             "incompleteDate", "incompleteTime", "URI"],
        "Parameter.DataType": ["text", "integer", "float", "date", "time", "datetime", "string", "boolean", "double",
                             "hexBinary", "base64Binary", "hexFloat", "base64Float", "partialDate", "partialTime",
                             "partialDatetime", "durationDatetime", "intervalDatetime", "incompleteDatetime",
                             "incompleteDate", "incompleteTime", "URI"],
        "CodeList.DataType": ["integer", "float", "text", "string"],
        "MethodDef.Type": ["Computation", "Imputation", "Transpose", "Other"],
        "StudyEventDef.Type": ["Scheduled", "Unscheduled", "Common"],
        "ODM.FileType": ["Snapshot", "Transactional"],
        "ODM.Granularity": ["All", "Metadata", "AdminData", "ReferenceData", "AllClinicalData", "SingleSite",
                            "SingleSubject"],
        "ODM.Archival": ["Yes", "No"],
        "ODM.ODMVersion": ["1.2", "1.2.1", "1.3", "1.3.1", "1.3.2", "2.0"],
        "User.UserType": ["Sponsor", "Investigator", "Lab", "Other"],
        "Location.LocationType": ["Sponsor", "Site", "CRO", "Lab", "Other"],
        "SignatureDef.Methodology": ["Digital", "Electronic"],
        "Country._content": ['AF', 'AL', 'DZ', 'AS', 'AD', 'AO', 'AI', 'AQ', 'AG', 'AR', 'AM', 'AW', 'AU', 'AT', 'AZ',
                             'BS', 'BH', 'BD', 'BB', 'BY', 'BE', 'BZ', 'BJ', 'BM', 'BT', 'BO', 'BO', 'BA', 'BW', 'BV',
                             'BR', 'IO', 'BN', 'BN', 'BG', 'BF', 'BI', 'KH', 'CM', 'CA', 'CV', 'KY', 'CF', 'TD', 'CL',
                             'CN', 'CX', 'CC', 'CO', 'KM', 'CG', 'CD', 'CK', 'CR', 'CI', 'CI', 'HR', 'CU', 'CY', 'CZ',
                             'DK', 'DJ', 'DM', 'DO', 'EC', 'EG', 'SV', 'GQ', 'ER', 'EE', 'ET', 'FK', 'FO', 'FJ', 'FI',
                             'FR', 'GF', 'PF', 'TF', 'GA', 'GM', 'GE', 'DE', 'GH', 'GI', 'GR', 'GL', 'GD', 'GP', 'GU',
                             'GT', 'GG', 'GN', 'GW', 'GY', 'HT', 'HM', 'VA', 'HN', 'HK', 'HU', 'IS', 'IN', 'ID', 'IR',
                             'IQ', 'IE', 'IM', 'IL', 'IT', 'JM', 'JP', 'JE', 'JO', 'KZ', 'KE', 'KI', 'KP', 'KR', 'KR',
                             'KW', 'KG', 'LA', 'LV', 'LB', 'LS', 'LR', 'LY', 'LY', 'LI', 'LT', 'LU', 'MO', 'MK', 'MG',
                             'MW', 'MY', 'MV', 'ML', 'MT', 'MH', 'MQ', 'MR', 'MU', 'YT', 'MX', 'FM', 'MD', 'MC', 'MN',
                             'ME', 'MS', 'MA', 'MZ', 'MM', 'MM', 'NA', 'NR', 'NP', 'NL', 'AN', 'NC', 'NZ', 'NI', 'NE',
                             'NG', 'NU', 'NF', 'MP', 'NO', 'OM', 'PK', 'PW', 'PS', 'PA', 'PG', 'PY', 'PE', 'PH', 'PN',
                             'PL', 'PT', 'PR', 'QA', 'RE', 'RO', 'RU', 'RU', 'RW', 'SH', 'KN', 'LC', 'PM', 'VC', 'VC',
                             'VC', 'WS', 'SM', 'ST', 'SA', 'SN', 'RS', 'SC', 'SL', 'SG', 'SK', 'SI', 'SB', 'SO', 'ZA',
                             'GS', 'SS', 'ES', 'LK', 'SD', 'SR', 'SJ', 'SZ', 'SE', 'CH', 'SY', 'TW', 'TW', 'TJ', 'TZ',
                             'TH', 'TL', 'TG', 'TK', 'TO', 'TT', 'TN', 'TR', 'TM', 'TC', 'TV', 'UG', 'UA', 'AE', 'GB',
                             'US', 'UM', 'UY', 'UZ', 'VU', 'VE', 'VE', 'VN', 'VN', 'VG', 'VI', 'WF', 'EH', 'YE', 'ZM',
                             'ZW']
    }

    @classmethod
    def value_set(cls, attribute):
        if attribute in cls._value_set:
            return cls._value_set[attribute]
        else:
            raise ValueError(f"Unknown value {attribute} in ValueSet. Unable to check value.")



import unittest
import odmlib.odm_1_3_2.model as ODM
import odmlib.typed as T
import odmlib.odm_element as OE


class TestText(OE.ODMElement):
    Name = T.String(required=True)
    OrderNumber = T.PositiveInteger(required=False)


class TestDescriptor(unittest.TestCase):

    def test_assignment(self):
        test = TestText(Name="test name", OrderNumber="1")
        self.assertEqual(test.Name, "test name")
        with self.assertRaises(TypeError):
            test.OID = None
        TestText.Name = "VariableOne"
        self.assertEqual(TestText.Name, "VariableOne")

    def test_get_missing_attribute(self):
        igd = ODM.ItemGroupDef(OID="IG.VS", Name="Vital Signs", Repeating="Yes")
        self.assertEqual(igd.OID, "IG.VS")
        self.assertIsNone(igd.Comment)

    def test_get_missing_element_with_required(self):
        itd = ODM.ItemDef(OID="IT.VS.VSORRES", Name="Vital Signs Results", DataType="text")
        self.assertEqual(itd.OID, "IT.VS.VSORRES")
        self.assertIsNone(itd.CodeListRef)

    def test_get_missing_undefined_attribute(self):
        itd = ODM.ItemDef(OID="IT.VS.VSORRES", Name="Vital Signs Results", DataType="text")
        self.assertEqual(itd.OID, "IT.VS.VSORRES")
        self.assertListEqual(itd.MeasurementUnitRef, [])
        result = itd.CodeListRef
        self.assertIsNone(result)
        itd.Alias.append(itd.CodeListRef)
        self.assertListEqual(itd.Alias, [None])
        with self.assertRaises(TypeError):
            itd.new_thing = "hello"




import unittest
import os
import odmlib.odm_parser as P
from xml.etree.ElementTree import Element
ODM_NS = "{http://www.cdisc.org/ns/odm/v1.3}"


class TestOdmParserMetaData(unittest.TestCase):
    def setUp(self) -> None:
        self.odm_file_1 = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'cdash-odm-test.xml')
        self.parser = P.ODMParser(self.odm_file_1)
        self.root = self.parser.parse()
        self.mdv = self.parser.MetaDataVersion()

    def test_MetaDataVersion(self):
        self.assertTrue(isinstance(self.mdv, list))
        self.assertDictEqual(self.mdv[0].attrib, {"Name": "TRACE-XML MDV", "OID": "MDV.TRACE-XML-ODM-01"})
        self.assertEqual(81, len([e.tag for e in self.mdv[0].getchildren()]))

    def test_Protocol(self):
        protocol = self.parser.Protocol(parent=self.mdv[0])
        self.assertIsInstance(protocol[0]["elem"], Element)
        self.assertListEqual([ODM_NS + "StudyEventRef"], [e.tag for e in protocol[0]["elem"].getchildren()])

    def test_StudyEventRef(self):
        protocol = self.parser.Protocol(parent=self.mdv[0])
        ser = self.parser.StudyEventRef(parent=protocol[0]["elem"])
        self.assertEqual(ser[0]["StudyEventOID"], "BASELINE")

    def test_StudyEventDef(self):
        sed = self.parser.StudyEventDef(parent=self.mdv[0])
        self.assertEqual(sed[0]["OID"], "BASELINE")
        self.assertIsInstance(sed[0]["elem"], Element)

    def test_FormRef(self):
        sed = self.parser.StudyEventDef(parent=self.mdv[0])
        fr = self.parser.FormRef(parent=sed[0]["elem"])
        self.assertEqual(fr[0]["FormOID"], "ODM.F.DM")
        fr_list = [ODM_NS + "FormRef", ODM_NS + "FormRef", ODM_NS + "FormRef"]
        self.assertListEqual(fr_list, [e.tag for e in sed[0]["elem"].getchildren()])

    def test_FormDef(self):
        fd = self.parser.FormDef(parent=self.mdv[0])
        self.assertEqual(fd[0]["OID"], "ODM.F.DM")
        self.assertEqual(fd[2]["Name"], "Adverse Event")
        self.assertIsInstance(fd[0]["elem"], Element)

    def test_ItemGroupDef(self):
        ig = self.parser.ItemGroupDef(parent=self.mdv[0])
        self.assertEqual(len(ig), 7)
        self.assertDictEqual({'Name': 'Demographics', 'OID': 'ODM.IG.DM', 'Repeating': 'No'}, ig[1]['elem'].attrib)
        self.assertEqual(ig[3]["OID"], "ODM.IG.VS")

    def test_ItemRef(self):
        ig = self.parser.ItemGroupDef(parent=self.mdv[0])
        ir = self.parser.ItemRef(parent=ig[0]['elem'])
        self.assertEqual(len(ir), 4)
        self.assertDictEqual({'ItemOID': 'ODM.IT.Common.SubjectID', 'Mandatory': 'Yes'}, ir[2]['elem'].attrib)





if __name__ == '__main__':
    unittest.main()



import odmlib.odm_element as OE
import odmlib.typed as T
import odmlib.ns_registry as NS

NS.NamespaceRegistry(prefix="odm", uri="http://www.cdisc.org/ns/odm/v2.0", is_default=True)
NS.NamespaceRegistry(prefix="xs", uri="http://www.w3.org/2001/XMLSchema-instance")
NS.NamespaceRegistry(prefix="xml", uri="http://www.w3.org/XML/1998/namespace")
NS.NamespaceRegistry(prefix="xlink", uri="http://www.w3.org/1999/xlink")


class TranslatedText(OE.ODMElement):
    lang = T.String(namespace="xml")
    Type = T.String()
    _content = T.String(required=True)


class Description(OE.ODMElement):
    TranslatedText = T.ODMListObject(required=True, element_class=TranslatedText)


class Alias(OE.ODMElement):
    Context = T.String(required=True)
    Name = T.String(required=True)


class StudyDescription(OE.ODMElement):
    Description = T.ODMObject(element_class=Description, required=True)


class ProtocolName(OE.ODMElement):
    _content = T.String(required=True)


class StudyName(OE.ODMElement):
    _content = T.String(required=True)


class Include(OE.ODMElement):
    StudyOID = T.OIDRef(required=True)
    MetaDataVersionOID = T.OIDRef(required=True)


class StudyEventRef(OE.ODMElement):
    StudyEventOID = T.OID(required=True)
    OrderNumber = T.Integer(required=False)
    Mandatory = T.ValueSetString(required=True)
    CollectionExceptionConditionOID = T.OIDRef()


class Protocol(OE.ODMElement):
    Description = T.ODMObject(element_class=Description)
    StudyEventRef = T.ODMListObject(element_class=StudyEventRef)
    Alias = T.ODMListObject(element_class=Alias)


class StudyEventDef(OE.ODMElement):
    """ represents ODM v2.0 StudyEventDef and can serialize as JSON or XML """
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    Repeating = T.ValueSetString(required=True)
    Type = T.ValueSetString(required=True)
    Category = T.String(required=False)
    Description = T.ODMObject(element_class=Description)
    ItemGroupRef = T.ODMListObject(element_class=ItemGroupRef)
    WorkflowRef = T.ODMListObject(element_class=WorkflowRef)
    Alias = T.ODMListObject(element_class=Alias)

    def __len__(self):
        """ returns the number of FormRefs in an StudyEventDef object as the length """
        return len(self.FormRef)

    def __getitem__(self, position):
        """
        creates an iterator from an StudyEventDef object that returns the FormRef in position
        """
        return self.FormRef[position]

    def __iter__(self):
        return iter(self.FormRef)


class ItemGroupRef(OE.ODMElement):
    ItemGroupOID = T.OID(required=True)
    OrderNumber = T.Integer(required=False)
    Mandatory = T.ValueSetString(required=True)
    CollectionExceptionConditionOID = T.OIDRef()


class ArchiveLayout(OE.ODMElement):
    OID = T.OID(required=True)
    PdfFileName = T.FileName(required=True)
    PresentationOID = T.OIDRef(required=False)


    def __len__(self):
        return len(self.ItemGroupRef)

    def __getitem__(self, position):
        return self.ItemGroupRef[position]

    def __iter__(self):
        return iter(self.ItemGroupRef)


class PDFPageRef(OE.ODMElement):
    Type = T.ValueSetString(required=True)
    PageRefs = T.String()
    FirstPage = T.PositiveInteger()
    LastPage = T.PositiveInteger()


class DocumentRef(OE.ODMElement):
    leafID = T.String(required=True)
    PDFPageRef = T.ODMListObject(element_class=PDFPageRef, namespace="def")


class SourceItem(OE.ODMElement):
    leadID = T.IDRef()
    ItemGroupOID = T.OIDRef()
    Resource = T.String()
    Attribute = T.String()
    Path = T.String()
    Label = T.String()


class SourceItems(OE.ODMElement):
    SourceItem = T.ODMListObject(element_class=SourceItem, required=True)


class Origin(OE.ODMElement):
    Type = T.ValueSetString(required=True)
    Source = T.ValueSetString()
    DocumentRef = T.ODMListObject(element_class=DocumentRef, namespace="def")
    Description = T.ODMObject(element_class=Description)
    SourceItems = T.ODMObject(element_class=SourceItems)


class ItemRef(OE.ODMElement):
    ItemOID = T.String(required=True)
    OrderNumber = T.Integer(required=False)
    Mandatory = T.ValueSetString(required=True)
    KeySequence = T.Integer(required=False)
    MethodOID = T.String(required=False)
    Role = T.String()
    RoleCodeListOID = T.String()
    CollectionExceptionConditionOID = T.String()


class ItemGroupDef(OE.ODMElement):
    """ represents ODM v2.0 ItemGroupDef and can serialize as JSON or XML"""
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    Repeating = T.ValueSetString(required=True)
    Type = T.ValueSetString()
    IsReferenceData = T.ValueSetString(required=False)
    DatasetName = T.Name()
    Domain = T.String()
    Purpose = T.String()
    CommentOID = T.OIDRef()
    Description = T.ODMObject(element_class=Description)
    ItemGroupRef = T.ODMListObject(element_class=ItemGroupRef)
    ItemRef = T.ODMListObject(element_class=ItemRef)
    WorkflowRef = T.ODMListObject(element_class=WorkflowRef)
    Origin = T.ODMListObject(element_class=Origin)
    Alias = T.ODMListObject(element_class=Alias)

    def __len__(self):
        return len(self.ItemRef)

    def __getitem__(self, position):
        return self.ItemRef[position]

    def __iter__(self):
        return iter(self.ItemRef)


class Question(OE.ODMElement):
    TranslatedText = T.ODMListObject(required=True, element_class=TranslatedText)


class ExternalQuestion(OE.ODMElement):
    Dictionary = T.String(required=False)
    Version = T.String(required=False)
    Code = T.String(required=False)


class MeasurementUnitRef(OE.ODMElement):
    MeasurementUnitOID = T.String(required=True)


class CheckValue(OE.ODMElement):
    _content = T.String(required=True)


class FormalExpression(OE.ODMElement):
    Context = T.String(required=True)
    _content = T.String(required=True)


class ErrorMessage(OE.ODMElement):
    TranslatedText = T.ODMListObject(required=True, element_class=TranslatedText)


class RangeCheck(OE.ODMElement):
    """ represents ODM v2.0 RangeCheck element that is a child of ItemDef and can serialize as JSON or XML """
    Comparator = T.ValueSetString(required=False)
    SoftHard = T.ValueSetString()
    CheckValue = T.ODMListObject(element_class=CheckValue)
    FormalExpression = T.ODMListObject(element_class=FormalExpression)
    ErrorMessage = T.ODMObject(element_class=ErrorMessage)


class CodeListRef(OE.ODMElement):
    CodeListOID = T.String("CodeListOID", required=True)


class ItemDef(OE.ODMElement):
    """ represents ODM v2.0 ItemDef and can serialize as JSON or XML - ordering of properties matters """
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    DataType = T.ValueSetString(required=True)
    Length = T.PositiveInteger()
    FractionDigits = T.NonNegativeInteger()
    DatasetVarName = T.Name()
    SDSVarName = T.SASName()
    CommentOID = T.String()
    Description = T.ODMObject(element_class=Description)
    Question = T.ODMObject(element_class=Question)
    ExternalQuestion = T.ODMObject(element_class=ExternalQuestion)
    MeasurementUnitRef = T.ODMListObject(element_class=MeasurementUnitRef)
    RangeCheck = T.ODMListObject(element_class=RangeCheck)
    CodeListRef = T.ODMObject(element_class=CodeListRef)
    Alias = T.ODMListObject(element_class=Alias)


class Decode(OE.ODMElement):
    TranslatedText = T.ODMListObject(required=True, element_class=TranslatedText)


class CodeListItem(OE.ODMElement):
    """ represents ODM CodeListItem element that is a child of CodeList and can serialize as JSON or XML """
    CodedValue = T.String(required=True)
    Rank = T.Float(required=False)
    OrderNumber = T.Integer(required=False)
    Decode = T.ODMObject(element_class=Decode)
    Alias = T.ODMListObject(element_class=Alias)


class EnumeratedItem(OE.ODMElement):
    """ represents ODM EnumeratedItem element that is a child of CodeList and can serialize as JSON or XML """
    CodedValue = T.String(required=True)
    Rank = T.Float(required=False)
    OrderNumber = T.Integer(required=False)
    Alias = T.ODMListObject(element_class=Alias)


class ExternalCodeList(OE.ODMElement):
    Dictionary = T.String(required=False)
    Version = T.String(required=False)
    ref = T.String(required=False)
    href = T.String(required=False)


class CodeList(OE.ODMElement):
    """ represents ODM v1.3.2 CodeList element that can serialize as JSON or XML """
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    DataType = T.ValueSetString(required=True)
    SASFormatName = T.SASFormat()
    Description = T.ODMObject(element_class=Description)
    CodeListItem = T.ODMListObject(element_class=CodeListItem)
    EnumeratedItem = T.ODMListObject(element_class=EnumeratedItem)
    ExternalCodeList = T.ODMObject(element_class=ExternalCodeList)
    Alias = T.ODMListObject(element_class=Alias)


class ConditionDef(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    Description = T.ODMObject(element_class=Description)
    FormalExpression = T.ODMListObject(element_class=FormalExpression)
    Alias = T.ODMListObject(element_class=Alias)


class Parameter(OE.ODMElement):
    Name = T.Name(required=True)
    Definition = T.String()
    DataType = T.ValueSetString(required=True)
    OrderNumber = T.PositiveInteger()


class ReturnValue(OE.ODMElement):
    Name = T.Name(required=True)
    Definition = T.String()
    DataType = T.ValueSetString(required=True)
    OrderNumber = T.PositiveInteger()


class MethodSignature(OE.ODMElement):
    Parameter = T.ODMListObject(element_class=Parameter)
    ReturnValue = T.ODMListObject(element_class=ReturnValue)


class MethodDef(OE.ODMElement):
    """ represents ODM v2.0 MethodDef and can serialize as JSON or XML """
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    Type = T.ValueSetString(required=True)
    Description = T.ODMObject(required=True, element_class=Description)
    MethodSignature = T.ODMObject(element_class=MethodSignature)
    FormalExpression = T.ODMListObject(element_class=FormalExpression)
    Alias = T.ODMListObject(element_class=Alias)


class StudyEventGroupRef(OE.ODMElement):
    StudyEventGroupOID = T.OIDRef(required=True)
    OrderNumber = T.Integer()
    Mandatory = T.ValueSetString(required=True)
    CollectionExceptionConditionOID = T.OIDRef()
    Description = T.ODMObject(element_class=Description)


class ExceptionEvent(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    ConditionOID = T.OIDRef(required=True)
    Description = T.ODMObject(element_class=Description)
    WorkflowRef = T.ODMObject(element_class=WorkflowRef)
    StudyEventGroupRef = T.ODMListObject(element_class=StudyEventGroupRef)
    StudyEventRef = T.ODMListObject(element_class=StudyEventRef)


class WorkflowRef(OE.ODMElement):
    WorkflowOID = T.OID(required=True)


class Arm(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    Description = T.ODMObject(element_class=Description)
    WorkflowRef = T.ODMObject(element_class=WorkflowRef)


class Epoch(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    SequenceNumber = T.PositiveInteger(required=True)
    Description = T.ODMObject(element_class=Description)


class StudyStructure(OE.ODMElement):
    Description = T.ODMObject(element_class=Description)
    Arm = T.ODMListObject(element_class=Arm)
    Epoch = T.ODMListObject(element_class=Epoch)
    WorkflowRef = T.ODMListObject(element_class=WorkflowRef)


class WorkflowStart(OE.ODMElement):
    StartOID = T.OIDRef(required=True)


class Transition(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    SourceOID = T.OIDRef(required=True)
    TargetOID = T.OIDRef(required=True)
    StartConditionOID = T.OIDRef()
    EndConditionOID = T.OIDRef()


class TargetTransition(OE.ODMElement):
    TargetTransitionOID = T.OIDRef(required=True)
    ConditionOID = T.OIDRef(required=True)


class DefaultTransition(OE.ODMElement):
    TargetTransitionOID = T.OIDRef(required=True)


class Branching(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    Type = T.ValueSetString(required=True)
    TargetTransition = T.ODMListObject(element_class=TargetTransition, required=True)
    DefaultTransition = T.ODMListObject(element_class=DefaultTransition)


class WorkflowEnd(OE.ODMElement):
    EndOID = T.OIDRef(required=True)


class WorkflowDef(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    Description = T.ODMObject(element_class=Description)
    WorkflowStart = T.ODMObject(element_class=WorkflowStart, required=True)
    Transition = T.ODMListObject(element_class=Transition)
    Branching = T.ODMListObject(element_class=Branching)
    WorkflowEnd = T.ODMListObject(element_class=WorkflowEnd, required=True)


class AbsoluteTimingConstraint(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    StudyEventGroupOID = T.OIDRef()
    StudyEventOID = T.OIDRef()
    TimepointTarget = T.IncompleteDateTimeString(required=True)
    TimepointPreWindow = T.DurationDateTimeString(required=True)
    TimepointPostWindow = T.DurationDateTimeString(required=True)
    Description = T.ODMObject(element_class=Description)


class RelativeTimingConstraint(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    PredecessorStudyEventGroupOID = T.OIDRef()
    PredecessorStudyEventOID = T.OIDRef()
    SuccessorStudyEventGroupOID = T.OIDRef()
    SuccessorStudyEventOID = T.OIDRef()
    Type = T.ValueSetString()
    TimepointRelativeTarget = T.DurationDateTimeString(required=True)
    TimepointPreWindow = T.DurationDateTimeString(required=True)
    TimepointPostWindow = T.DurationDateTimeString(required=True)
    Description = T.ODMObject(element_class=Description)


class TransitionTimingConstraint(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    TransitionOID = T.OIDRef(required=True)
    TimepointRelativeTarget = T.DurationDateTimeString(required=True)
    MethodOID = T.OIDRef()
    TimepointPreWindow = T.DurationDateTimeString(required=True)
    TimepointPostWindow = T.DurationDateTimeString(required=True)
    Description = T.ODMObject(element_class=Description)


class DurationTimingConstraint(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    StructuralElementOID = T.OIDRef(required=True)
    DurationTarget = T.DurationDateTimeString(required=True)
    DurationPreWindow = T.DurationDateTimeString(required=True)
    DurationPostWindow = T.DurationDateTimeString(required=True)
    Description = T.ODMObject(element_class=Description)


class StudyTiming(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    AbsoluteTimingConstraint = T.ODMListObject(element_class=AbsoluteTimingConstraint)
    RelativeTimingConstraint = T.ODMListObject(element_class=RelativeTimingConstraint)
    TransitionTimingConstraint = T.ODMObject(element_class=TransitionTimingConstraint)
    DurationTimingConstraint = T.ODMListObject(element_class=DurationTimingConstraint)


class StudyEventGroupDef(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    ArmOID = T.OIDRef(required=True)
    EpochOID = T.OIDRef(required=True)
    Description = T.ODMObject(element_class=Description)


class MetaDataVersion(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    Description = T.ODMObject(element_class=Description)
    Include = T.ODMObject(element_class=Include)
    Protocol = T.ODMObject(element_class=Protocol)
    StudyStructure = T.ODMObject(element_class=StudyStructure)
    WorkflowDef = T.ODMListObject(element_class=WorkflowDef)
    StudyTiming = T.ODMObject(element_class=StudyTiming)
    StudyEventGroupDef = T.ODMListObject(element_class=StudyEventGroupDef)
    StudyEventDef = T.ODMListObject(element_class=StudyEventDef)
    ItemGroupDef = T.ODMListObject(element_class=ItemGroupDef)
    ItemDef = T.ODMListObject(element_class=ItemDef)
    CodeList = T.ODMListObject(element_class=CodeList)
    ConditionDef = T.ODMListObject(element_class=ConditionDef)
    MethodDef = T.ODMListObject(element_class=MethodDef)


class LoginName(OE.ODMElement):
    _content = T.String(required=True)


class DisplayName(OE.ODMElement):
    _content = T.String(required=True)


class FullName(OE.ODMElement):
    _content = T.String(required=True)


class FirstName(OE.ODMElement):
    _content = T.String(required=True)


class LastName(OE.ODMElement):
    _content = T.String(required=True)


class Organization(OE.ODMElement):
    _content = T.String(required=True)


class StreetName(OE.ODMElement):
    _content = T.String(required=True)


class City(OE.ODMElement):
    _content = T.String(required=True)


class StateProv(OE.ODMElement):
    _content = T.String(required=True)


class Country(OE.ODMElement):
    _content = T.ValueSetString(required=True)


class PostalCode(OE.ODMElement):
    _content = T.String(required=True)


class OtherText(OE.ODMElement):
    _content = T.String(required=True)


class Address(OE.ODMElement):
    StreetName = T.ODMListObject(element_class=StreetName)
    City = T.ODMObject(element_class=City)
    StateProv = T.ODMObject(element_class=StateProv)
    Country = T.ODMObject(element_class=Country)
    PostalCode = T.ODMObject(element_class=PostalCode)
    OtherText = T.ODMObject(element_class=OtherText)


class Email(OE.ODMElement):
    _content = T.Email(required=True)


class Picture(OE.ODMElement):
    PictureFileName = T.FileName(required=True)
    ImageType = T.Name()


class Pager(OE.ODMElement):
    _content = T.String(required=True)


class Fax(OE.ODMElement):
    _content = T.String(required=True)


class Phone(OE.ODMElement):
    _content = T.String(required=True)


class LocationRef(OE.ODMElement):
    LocationOID = T.OIDRef(required=True)


class Certificate(OE.ODMElement):
    _content = T.String(required=True)


class User(OE.ODMElement):
    OID = T.OID(required=True)
    UserType = T.ValueSetString()
    LoginName = T.ODMObject(element_class=LoginName)
    DisplayName = T.ODMObject(element_class=DisplayName)
    FullName = T.ODMObject(element_class=FullName)
    FirstName = T.ODMObject(element_class=FirstName)
    LastName = T.ODMObject(element_class=LastName)
    Organization = T.ODMObject(element_class=Organization)
    Address = T.ODMListObject(element_class=Address)
    Email = T.ODMListObject(element_class=Email)
    Pager = T.ODMObject(element_class=Pager)
    Fax = T.ODMListObject(element_class=Fax)
    Phone = T.ODMListObject(element_class=Phone)
    LocationRef = T.ODMListObject(element_class=LocationRef)
    Certificate = T.ODMListObject(element_class=Certificate)


class MetaDataVersionRef(OE.ODMElement):
    StudyOID = T.OIDRef(required=True)
    MetaDataVersionOID = T.OIDRef(required=True)
    EffectiveDate = T.DateString(required=True)


class Location(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    LocationType = T.ValueSetString()
    MetaDataVersionRef = T.ODMListObject(required=True, element_class=MetaDataVersionRef)


class Meaning(OE.ODMElement):
    _content = T.String(required=True)


class LegalReason(OE.ODMElement):
    _content = T.String(required=True)


class SignatureDef(OE.ODMElement):
    OID = T.OID(required=True)
    Methodology = T.ValueSetString()
    Meaning = T.ODMObject(required=True, element_class=Meaning)
    LegalReason = T.ODMObject(required=True, element_class=LegalReason)


class AdminData(OE.ODMElement):
    StudyOID = T.OIDRef()
    User = T.ODMListObject(element_class=User)
    Location = T.ODMListObject(element_class=Location)
    SignatureDef = T.ODMListObject(element_class=SignatureDef)


class Study(OE.ODMElement):
    """ v2.0 """
    OID = T.String(required=True)
    StudyName = T.String(required=True)
    ProtocolName = T.String(required=True)
    Description = T.ODMObject(required=True, element_class=Description)
    MetaDataVersion = T.ODMListObject(required=False, element_class=MetaDataVersion)


class ODM(OE.ODMElement):
    """ v2.0 """
    Description = T.String(required=False)
    FileType = T.ValueSetString(required=True)
    Granularity = T.ValueSetString(required=False)
    Archival = T.ValueSetString(required=False)
    FileOID = T.OID(required=True)
    CreationDateTime = T.DateTimeString(required=True)
    PriorFileOID = T.OIDRef(required=False)
    AsOfDateTime = T.DateTimeString(required=False)
    ODMVersion = T.ValueSetString(required=False)
    Originator = T.String(required=False)
    SourceSystem = T.String(required=False)
    SourceSystemVersion = T.String(required=False)
    ID = T.ID()
    Study = T.ODMListObject(element_class=Study)
    AdminData = T.ODMListObject(element_class=AdminData)
    #ReferenceData = T.ODMListObject(element_class=referencedata.ReferenceData)
    #ClinicalData = T.ODMListObject(element_class=clinicaldata.ClinicalData)
    #Association = T.ODMListObject(element_class=association.Association)
    #ds_Signature = T.ODMListObject(element_class=dssignature.ds_Signature)




from unittest import TestCase
import json
import odmlib.odm_1_3_2.model as ODM


class TestItemDef(TestCase):
    def setUp(self) -> None:
        attrs = self.set_item_attributes()
        self.item = ODM.ItemDef(**attrs)

    def test_required_attributes_only(self):
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text"}
        self.item = ODM.ItemDef(**attrs)
        self.assertEqual(self.item.OID, "ODM.IT.AE.AEYN")

    def test_add_range_check(self):
        rc = ODM.RangeCheck(Comparator="EQ", SoftHard="Soft")
        tt1 = ODM.TranslatedText(_content="invalid test code", lang="en")
        tt2 = ODM.TranslatedText(_content="code de test invalide", lang="fr")
        rc.CheckValue = [ODM.CheckValue(_content="DIABP")]
        rc.FormalExpression = [ODM.FormalExpression(Context="Python 3.7", _content="print('hello world')")]
        rc.MeasurementUnitRef = ODM.MeasurementUnitRef(MeasurementUnitOID="ODM.MU.UNITS")
        attrs = {"Comparator": "GT", "SoftHard": "Soft"}
        rc1 = ODM.RangeCheck(**attrs)
        tt1 = ODM.TranslatedText(_content="invalid test code", lang="en")
        rc1.ErrorMessage.TranslatedText = [tt1]
        rc1.CheckValue = [ODM.CheckValue(_content="SYSBP")]
        rc1.FormalExpression = [ODM.FormalExpression(Context="Python 3.7", _content="print('hello world')")]
        rc1.MeasurementUnitRef = ODM.MeasurementUnitRef(MeasurementUnitOID="ODM.MU.UNITS")
        self.item.RangeCheck = [rc, rc1]
        self.assertEqual(self.item.RangeCheck[0].Comparator, "EQ")
        self.assertEqual(self.item.RangeCheck[1].Comparator, "GT")

    def test_set_description(self):
        attrs = {"_content": "this is the first test description", "lang": "en"}
        tt1 = ODM.TranslatedText(**attrs)
        self.item.Description.TranslatedText.append(tt1)
        self.assertEqual(self.item.Description.TranslatedText[0].lang, "en")
        self.assertEqual(self.item.Description.TranslatedText[0]._content, "this is the first test description")

    def test_set_invalid_description(self):
        rc = ODM.RangeCheck(Comparator="EQ", SoftHard="Soft")
        tt1 = ODM.TranslatedText(_content="invalid test code", lang="en")
        tt2 = ODM.TranslatedText(_content="code de test invalide", lang="fr")
        rc.ErrorMessage.TranslatedText.append(tt1)
        rc.ErrorMessage.TranslatedText.append(tt2)
        rc.CheckValue = [ODM.CheckValue(_content="DIABP")]
        rc.FormalExpression = [ODM.FormalExpression(Context="Python 3.7", _content="print('hello world')")]
        rc.MeasurementUnitRef = [ODM.MeasurementUnitRef(MeasurementUnitOID="ODM.MU.UNITS")]
        self.item.RangeCheck = [rc]
        # Description requires a Description object, not a RangeCheck object
        with self.assertRaises(TypeError):
            self.item.Description = rc

    def test_add_description(self):
        attrs = {"_content": "this is the first test description", "lang": "en"}
        tt1 = ODM.TranslatedText(**attrs)
        attrs = {"_content": "this is the second test description", "lang": "en"}
        tt2 = ODM.TranslatedText(**attrs)
        self.item.Description.TranslatedText.append(tt1)
        self.item.Description.TranslatedText.append(tt2)
        self.assertEqual(self.item.Description.TranslatedText[1]._content, "this is the second test description")

    def test_add_question(self):
        tt1 = ODM.TranslatedText(_content="Any AEs?", lang="en")
        self.item.Question.TranslatedText.append(tt1)
        self.assertEqual(self.item.Question.TranslatedText[0]._content, "Any AEs?")

    def test_external_question(self):
        self.item.ExternalQuestion = ODM.ExternalQuestion(Dictionary="SF36", Version="12", Code="Walks 1-mile")
        self.assertEqual(self.item.ExternalQuestion.Dictionary, "SF36")

    def test_add_measurement_unit_ref(self):
        self.item.MeasurementUnitRef.append(ODM.MeasurementUnitRef(MeasurementUnitOID="MU.UNITS"))
        self.item.MeasurementUnitRef.append(ODM.MeasurementUnitRef(MeasurementUnitOID="MU2.UNITS2"))
        self.assertEqual(self.item.MeasurementUnitRef[0].MeasurementUnitOID, "MU.UNITS")

    def test_codelist_ref(self):
        self.item.CodeListRef = ODM.CodeListRef(CodeListOID="CL.NY_SUB_Y_N_2011-10-24")
        self.assertEqual(self.item.CodeListRef.CodeListOID, "CL.NY_SUB_Y_N_2011-10-24")

    def test_codelist_ref_exists_check(self):
        attrs = self.set_item_attributes()
        item = ODM.ItemDef(**attrs)
        codelistref_check_succeeds = False
        if not item.CodeListRef:
            codelistref_check_succeeds = True
        if item.CodeListRef:
            codelistref_check_succeeds = False
        self.assertTrue(codelistref_check_succeeds)

    def test_add_alias(self):
        self.item.Alias = [ODM.Alias(Context="CDASH", Name="AEYN")]
        self.assertEqual(self.item.Alias[0].Name, "AEYN")

    def test_to_json(self):
        attrs = self.set_item_attributes()
        item = ODM.ItemDef(**attrs)
        item.Description.TranslatedText.append(ODM.TranslatedText(_content="this is the first test description", lang="en"))
        item.Question.TranslatedText = [ODM.TranslatedText(_content="Any AEs?", lang="en")]
        item.CodeListRef = ODM.CodeListRef(CodeListOID="CL.NY_SUB_Y_N_2011-10-24")
        item_json = item.to_json()
        item_dict = json.loads(item_json)
        self.assertEqual(item_dict["OID"], "ODM.IT.AE.AEYN")

    def test_to_dict(self):
        attrs = self.set_item_attributes()
        item = ODM.ItemDef(**attrs)
        item.Description.TranslatedText = [ODM.TranslatedText(_content="this is the first test description", lang="en")]
        item.Question.TranslatedText = [ODM.TranslatedText(_content="Any AEs?", lang="en")]
        item.CodeListRef = ODM.CodeListRef(CodeListOID="CL.NY_SUB_Y_N_2011-10-24")
        item_dict = item.to_dict()
        expected_dict = {'OID': 'ODM.IT.AE.AEYN', 'Name': 'Any AEs?', 'DataType': 'text', 'Length': 1,
                         'SASFieldName': 'AEYN', 'SDSVarName': 'AEYN', 'Origin': 'CRF',
                         'Comment': 'Data management field',
                         'Description': {'TranslatedText': [{'_content': 'this is the first test description',
                                          'lang': 'en'}]},
                         'Question': {'TranslatedText': [{'_content': 'Any AEs?', 'lang': 'en'}]},
                         'CodeListRef': {'CodeListOID': 'CL.NY_SUB_Y_N_2011-10-24'}}
        self.assertDictEqual(item_dict, expected_dict)

    def test_to_dict_description(self):
        tt1 = ODM.TranslatedText(_content="this is the first test description", lang="en")
        tt2 = ODM.TranslatedText(_content="this is the second test description", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt1, tt2]
        desc_dict = desc.to_dict()
        print(desc_dict)
        self.assertDictEqual(desc_dict["TranslatedText"][1],
                             {'_content': 'this is the second test description', 'lang': 'en'})

    def test_to_xml_description(self):
        tt1 = ODM.TranslatedText(_content="this is the first test description", lang="en")
        tt2 = ODM.TranslatedText(_content="this is the second test description", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt1, tt2]
        desc_xml = desc.to_xml()
        tts = desc_xml.findall("TranslatedText")
        self.assertEqual(tts[0].text, "this is the first test description")

    def test_to_xml(self):
        attrs = self.set_item_attributes()
        item = ODM.ItemDef(**attrs)
        item.Description.TranslatedText = [ODM.TranslatedText(_content="this is the first test description", lang="en")]
        item.Question.TranslatedText = [ODM.TranslatedText(_content="Any AEs?", lang="en")]
        item.ExternalQuestion = ODM.ExternalQuestion(Dictionary="SF36", Version="12", Code="Walks 1-mile")
        item.CodeListRef = ODM.CodeListRef(CodeListOID="CL.NY_SUB_Y_N_2011-10-24")
        item.RangeCheck = [ODM.RangeCheck(Comparator="EQ", SoftHard="Soft")]
        item.RangeCheck[0].CheckValue = [ODM.CheckValue(_content="DIABP")]
        item.RangeCheck[0].FormalExpression = [ODM.FormalExpression(Context="Python 3.7", _content="print('hello world')")]
        item.RangeCheck[0].MeasurementUnitRef = ODM.MeasurementUnitRef(MeasurementUnitOID="ODM.MU.UNITS")
        item.RangeCheck[0].ErrorMessage.TranslatedText = [ODM.TranslatedText(_content="invalid test code", lang="en")]
        item_xml = item.to_xml()
        self.assertEqual(item_xml.attrib["OID"], "ODM.IT.AE.AEYN")
        cv = item_xml.find("*/CheckValue")
        self.assertEqual(cv.text, "DIABP")
        dt = item_xml.findall("Description/TranslatedText")
        self.assertEqual(len(dt), 1)


    def test_missing_itemdef_attributes(self):
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?"}
        with self.assertRaises(ValueError):
            ODM.ItemDef(**attrs)

    def test_invalid_attribute_data_type(self):
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "SignificantDigits": "A"}
        with self.assertRaises(TypeError):
            self.item = ODM.ItemDef(**attrs)

    def set_item_attributes(self):
        """
        set some ItemDef element attributes using test data
        :return: dictionary with ItemDef attribute information
        """
        return {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": "Data management field"}





OID_SKIP_LIST = ["FileOID"]

class OIDRef:
    def __init__(self):
        self.oid = {}
        self.oid_ref = {}
        self._init_oid_ref()
        self.ref_def = {}
        self._init_ref_def()
        self.def_ref = {}
        self._init_def_ref()

    def add_oid(self, oid, element):
        """ ODMLIB expects all OIDs to be unique within the scope of an ODM document """
        if oid in self.oid:
            raise ValueError(f"OID {oid} is not unique - element {element}")
        self.oid[oid] = element

    def add_oid_ref(self, oid, attr):
        self.oid_ref[attr].add(oid)

    def check_oid_refs(self):
        for attr, oid_set in self.oid_ref.items():
            for oid in oid_set:
                if attr not in OID_SKIP_LIST:
                    if oid not in self.oid:
                        raise ValueError(f"OID {oid} referenced in the attribute {attr} is not found.")
                    elif self.ref_def.get(attr) != self.oid.get(oid):
                        raise ValueError(f"OID reference for attribute {attr} element types do not match: "
                                         f"{self.ref_def.get(attr)} and {self.oid.get(oid)}")
        return True

    def check_unreferenced_oids(self):
        """ identify ELEMENTS that are defined but not used """
        orphans = {}
        for oid, elem in self.oid.items():
            for ref in self.def_ref[elem]:
                if oid not in self.oid_ref[ref]:
                    orphans[oid] = ref
        return orphans

    def _init_oid_ref(self):
        self.oid_ref["MetaDataVersionOID"] = set()
        self.oid_ref["StudyOID"] = set()
        self.oid_ref["StudyEventOID"] = set()
        self.oid_ref["FormOID"] = set()
        self.oid_ref["ItemGroupOID"] = set()
        self.oid_ref["ItemOID"] = set()
        self.oid_ref["MethodOID"] = set()
        self.oid_ref["CollectionExceptionConditionOID"] = set()
        self.oid_ref["PresentationOID"] = set()
        self.oid_ref["RoleCodeListOID"] = set()
        self.oid_ref["MeasurementUnitOID"] = set()
        self.oid_ref["CodeListOID"] = set()
        self.oid_ref["LocationOID"] = set()
        self.oid_ref["FileOID"] = set()
        self.oid_ref["PriorFileOID"] = set()
        self.oid_ref["ArchiveLayoutOID"] = set()
        self.oid_ref["WhereClauseOID"] = set()
        self.oid_ref["ValueListOID"] = set()
        self.oid_ref["CommentOID"] = set()

    def _init_ref_def(self):
        self.ref_def["MetaDataVersionOID"] = "MetaDataVersion"
        self.ref_def["StudyOID"] = "Study"
        self.ref_def["ItemGroupOID"] = "ItemGroupDef"
        self.ref_def["ItemOID"] = "ItemDef"
        self.ref_def["MethodOID"] = "MethodDef"
        self.ref_def["CodeListOID"] = "CodeList"
        self.ref_def["FileOID"] = "ODM"
        self.ref_def["PriorFileOID"] = "ODM"
        self.ref_def["LocationOID"] = "Location"
        self.ref_def["WhereClauseOID"] = "WhereClauseDef"
        self.ref_def["ValueListOID"] = "ValueListDef"
        self.ref_def["CommentOID"] = "CommentDef"

    def _init_def_ref(self):
        self.def_ref["MetaDataVersion"] = ["MetaDataVersionOID"]
        self.def_ref["Study"] = ["StudyOID"]
        self.def_ref["ItemGroupDef"] = ["ItemGroupOID"]
        self.def_ref["ItemDef"] = ["ItemOID"]
        self.def_ref["MethodDef"] = ["MethodOID"]
        self.def_ref["CodeList"] = ["CodeListOID"]
        self.def_ref["ConditionDef"] = ["CollectionExceptionConditionOID"]
        self.def_ref["Presentation"] = ["PresentationOID"]
        self.def_ref["ODM"] = ["FileOID", "PriorFileOID"]
        self.def_ref["Location"] = ["LocationOID"]
        self.def_ref["WhereClauseDef"] = ["WhereClauseOID"]
        self.def_ref["ValueListDef"] = ["ValueListOID"]
        self.def_ref["CommentDef"] = ["CommentOID"]



from abc import ABC, abstractmethod


class DocumentLoader(ABC):
    @abstractmethod
    def load_document(self, doc):
        raise NotImplementedError("Attempted to execute an abstract method load_document in the DocumentLoader class")

    @abstractmethod
    def create_document(self, filename):
        raise NotImplementedError("Attempted to execute an abstract method create_document in the DocumentLoader class")

    @abstractmethod
    def create_document_from_string(self, odm_string):
        raise NotImplementedError("Attempted to execute an abstract method create_document_from_string in the DocumentLoader class")

    @abstractmethod
    def load_metadataversion(self, idx):
        raise NotImplementedError("Attempted to execute an abstract method load_metadataversion in the DocumentLoader class")

    @abstractmethod
    def load_study(self, idx):
        raise NotImplementedError("Attempted to execute an abstract method load_study in the DocumentLoader class")

    @abstractmethod
    def load_odm(self):
        raise NotImplementedError("Attempted to execute an abstract method load_odm in the DocumentLoader class")



import unittest
import os
import odmlib.odm_parser as P
import odmlib.ns_registry as NS

ODM_NS = "{http://www.cdisc.org/ns/odm/v1.3}"


class TestOdmParserMetaData(unittest.TestCase):
    def setUp(self) -> None:
        self.odm_file_1 = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'define2-0-0-sdtm-test.xml')
        NS.NamespaceRegistry(prefix="odm", uri="http://www.cdisc.org/ns/odm/v1.3", is_default=True)
        NS.NamespaceRegistry(prefix="def", uri="http://www.cdisc.org/ns/def/v2.0")
        self.nsr = NS.NamespaceRegistry(prefix="nciodm", uri="http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC")
        self.parser = P.ODMParser(self.odm_file_1, self.nsr)
        self.root = self.parser.parse()
        self.mdv = self.parser.MetaDataVersion()

    def test_MetaDataVersion(self):
        self.assertTrue(isinstance(self.mdv, list))
        # elementree does not support using the suffix for accessing attributes not in the default ns
        mdv_dict = {"OID": "MDV.CDISC01.SDTMIG.3.1.2.SDTM.1.2", "Name": "Study CDISC01, Data Definitions",
                    "Description": "Study CDISC01, Data Definitions",
                    "{http://www.cdisc.org/ns/def/v2.0}DefineVersion": "2.0.0",
                    "{http://www.cdisc.org/ns/def/v2.0}StandardName": "SDTM-IG",
                    "{http://www.cdisc.org/ns/def/v2.0}StandardVersion": "3.1.2"}
        self.assertDictEqual(self.mdv[0].attrib, mdv_dict)
        self.assertEqual(mdv_dict["{http://www.cdisc.org/ns/def/v2.0}DefineVersion"],
                         self.mdv[0].attrib["{http://www.cdisc.org/ns/def/v2.0}DefineVersion"])
        self.assertEqual(769, len([e.tag for e in self.mdv[0].getchildren()]))

    def test_ItemGroupDef(self):
        ig = self.parser.ItemGroupDef(parent=self.mdv[0])
        self.assertEqual(len(ig), 34)
        dm_itg = {"OID": "IG.DM", "Domain": "DM", "Name": "DM", "Repeating": "No", "IsReferenceData": "No",
                  "SASDatasetName": "DM", "Purpose": "Tabulation",
                  "{http://www.cdisc.org/ns/def/v2.0}Structure": "One record per subject",
                  "{http://www.cdisc.org/ns/def/v2.0}Class": "SPECIAL PURPOSE",
                  "{http://www.cdisc.org/ns/def/v2.0}CommentOID": "COM.DOMAIN.DM",
                  "{http://www.cdisc.org/ns/def/v2.0}ArchiveLocationID": "LF.DM"}
        self.assertDictEqual(dm_itg, ig[5]['elem'].attrib)
        self.assertEqual(ig[4]["{http://www.cdisc.org/ns/def/v2.0}ArchiveLocationID"], "LF.TV")

    def test_ItemGroupDef_nsr(self):
        ig = self.parser.ItemGroupDef(parent=self.mdv[0])
        self.assertEqual(len(ig), 34)
        dm_itg = {"OID": "IG.DM", "Domain": "DM", "Name": "DM", "Repeating": "No", "IsReferenceData": "No",
                  "SASDatasetName": "DM", "Purpose": "Tabulation",
                  self.nsr.get_ns_attribute_name("Structure", "def"): "One record per subject",
                  self.nsr.get_ns_attribute_name("Class", "def"): "SPECIAL PURPOSE",
                  self.nsr.get_ns_attribute_name("CommentOID", "def"): "COM.DOMAIN.DM",
                  self.nsr.get_ns_attribute_name("ArchiveLocationID", "def"): "LF.DM"}
        self.assertDictEqual(dm_itg, ig[5]['elem'].attrib)
        self.assertEqual(ig[4][self.nsr.get_ns_attribute_name("ArchiveLocationID", "def")], "LF.TV")

    def test_ItemRef(self):
        ig = self.parser.ItemGroupDef(parent=self.mdv[0])
        ir = self.parser.ItemRef(parent=ig[0]['elem'])
        self.assertEqual(len(ir), 10)
        self.assertDictEqual({'ItemOID': 'IT.TA.ARMCD', 'Mandatory': 'Yes', "OrderNumber": "3", "KeySequence": "2"},
                             ir[2]['elem'].attrib)

    def test_ValueListDef(self):
        # test using an element from the define-xml v2.0 namespace
        vld = self.parser.ValueListDef(parent=self.mdv[0], ns_prefix="def")
        self.assertEqual(len(vld), 19)
        self.assertDictEqual({'OID': 'VL.DA.DAORRES'}, vld[0]['elem'].attrib)
        # now back to retrieving an element from the Define-XML doc that exists in the ODM namespace
        ir_list = self.parser.ItemRef(parent=vld[0]['elem'])
        self.assertEqual(len(ir_list), 2)
        wc = self.parser.WhereClauseRef(parent=ir_list[0]['elem'], ns_prefix="def")
        self.assertEqual(wc[0]["WhereClauseOID"], "WC.DA.DATESTCD.DISPAMT")


if __name__ == '__main__':
    unittest.main()



import unittest
import odmlib.odm_1_3_2.model as ODM


class TestAdminData(unittest.TestCase):
    def test_User(self):
        user = ODM.User(OID="ODM.U.3245", UserType="Sponsor")
        user.FirstName = ODM.FirstName(_content="John")
        user.LastName = ODM.LastName(_content="Rockerfeller")
        user.Organization = ODM.Organization(_content="Cleveland Clinic")
        user.Email.append(ODM.Email(_content="jrocker@clevelandclinic.org"))
        self.assertEqual(user.Email[0]._content, "jrocker@clevelandclinic.org")


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase
import json
import odmlib.odm_1_3_2.model as ODM


class TestStudyEventDef(TestCase):
    def setUp(self) -> None:
        attrs = self.set_attributes()
        self.sed = ODM.StudyEventDef(**attrs)

    def test_add_description(self):
        tt1 = ODM.TranslatedText(_content="this is the first test description", lang="en")
        tt2 = ODM.TranslatedText(_content="this is the second test description", lang="en")
        self.sed.Description = ODM.Description()
        self.sed.Description.TranslatedText = [tt1, tt2]
        self.assertEqual(len(self.sed.Description.TranslatedText), 2)
        self.assertEqual(self.sed.Description.TranslatedText[1]._content, 'this is the second test description')

    def test_add_bad_attribute(self):
        attrs = self.set_attributes()
        attrs["Type"] = "Rescheduled"
        with self.assertRaises(ValueError):
            self.sed = ODM.StudyEventDef(**attrs)

    def test_add_form_ref(self):
        fr1 = ODM.FormRef(FormOID="ODM.F.VS", Mandatory="Yes", OrderNumber=1)
        fr2 = ODM.FormRef(FormOID="ODM.F.DM", Mandatory="Yes", OrderNumber=2)
        fr3 = ODM.FormRef(FormOID="ODM.F.MH", Mandatory="Yes", OrderNumber=3)
        self.sed.FormRef = [fr1, fr2, fr3]
        self.assertEqual(self.sed.FormRef[0].FormOID, "ODM.F.VS")
        self.assertEqual(self.sed.FormRef[2].OrderNumber, 3)

    def test_add_alias(self):
        a1 = ODM.Alias(Context="SDTMIG", Name="VS")
        a2 = ODM.Alias(Context="CDASHIG", Name="VS")
        self.sed.Alias = [a1, a2]
        self.assertEqual(len(self.sed.Alias), 2)
        self.assertEqual(self.sed.Alias[1].Context, "CDASHIG")

    def test_to_json(self):
        attrs = self.set_attributes()
        sed = ODM.StudyEventDef(**attrs)
        tt1 = ODM.TranslatedText(_content="this is the first test description", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt1]
        sed.Description = desc
        fr1 = ODM.FormRef(FormOID="ODM.F.VS", Mandatory="Yes", OrderNumber=1)
        fr2 = ODM.FormRef(FormOID="ODM.F.DM", Mandatory="Yes", OrderNumber=2)
        fr3 = ODM.FormRef(FormOID="ODM.F.MH", Mandatory="Yes", OrderNumber=3)
        sed.FormRef = [fr1, fr2, fr3]
        a1 = ODM.Alias(Context="SDTMIG", Name="VS")
        sed.Alias = [a1]
        sed_json = sed.to_json()
        sed_dict = json.loads(sed_json)
        print(sed_dict)
        self.assertEqual(sed_dict["OID"], "ODM.SE.BASELINE")
        self.assertDictEqual(sed_dict, self.expected_dict())

    def test_to_dict(self):
        attrs = self.set_attributes()
        sed = ODM.StudyEventDef(**attrs)
        tt1 = ODM.TranslatedText(_content="this is the first test description", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt1]
        sed.Description = desc
        fr1 = ODM.FormRef(FormOID="ODM.F.VS", Mandatory="Yes", OrderNumber=1)
        fr2 = ODM.FormRef(FormOID="ODM.F.DM", Mandatory="Yes", OrderNumber=2)
        fr3 = ODM.FormRef(FormOID="ODM.F.MH", Mandatory="Yes", OrderNumber=3)
        sed.FormRef = [fr1, fr2, fr3]
        a1 = ODM.Alias(Context="SDTMIG", Name="VS")
        sed.Alias = [a1]
        sed_dict = sed.to_dict()
        self.assertEqual(sed_dict["OID"], "ODM.SE.BASELINE")
        self.assertDictEqual(sed_dict, self.expected_dict())

    def test_to_xml(self):
        attrs = self.set_attributes()
        sed = ODM.StudyEventDef(**attrs)
        tt1 = ODM.TranslatedText(_content="this is the first test description", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt1]
        sed.Description = desc
        fr1 = ODM.FormRef(FormOID="ODM.F.VS", Mandatory="Yes", OrderNumber=1)
        fr2 = ODM.FormRef(FormOID="ODM.F.DM", Mandatory="Yes", OrderNumber=2)
        fr3 = ODM.FormRef(FormOID="ODM.F.MH", Mandatory="Yes", OrderNumber=3)
        sed.FormRef = [fr1, fr2, fr3]
        a1 = ODM.Alias(Context="SDTMIG", Name="VS")
        sed.Alias = [a1]
        sed_xml = sed.to_xml()
        self.assertEqual(sed_xml.attrib["OID"], "ODM.SE.BASELINE")
        fr = sed_xml.findall("FormRef")
        self.assertEqual(len(fr), 3)
        self.assertEqual(fr[0].attrib, {"FormOID": "ODM.F.VS", "Mandatory": "Yes", "OrderNumber": "1"})

    def test_studyeventdef_slice(self):
        """ test the ability to reference a specific or slice of FormRefs """
        attrs = self.set_attributes()
        sed = ODM.StudyEventDef(**attrs)
        tt1 = ODM.TranslatedText(_content="this is the first test description", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt1]
        sed.Description = desc
        fr1 = ODM.FormRef(FormOID="ODM.F.VS", Mandatory="Yes", OrderNumber=1)
        fr2 = ODM.FormRef(FormOID="ODM.F.DM", Mandatory="Yes", OrderNumber=2)
        fr3 = ODM.FormRef(FormOID="ODM.F.MH", Mandatory="Yes", OrderNumber=3)
        sed.FormRef = [fr1, fr2, fr3]
        self.assertEqual(len(sed.FormRef), 3)
        first_fr = sed[0]
        self.assertEqual(first_fr.FormOID, "ODM.F.VS")
        slice_fr = sed[1:3]
        self.assertEqual(slice_fr[1].FormOID, "ODM.F.MH")

    @staticmethod
    def set_attributes():
        """
        set some StudyEventDef element attributes using test data
        :return: dictionary with StudyEventDef attribute information
        """
        return {"OID": "ODM.SE.BASELINE", "Name": "Baseline Visit", "Repeating": "No", "Type": "Scheduled",
                "Category": "Pre-treatment"}

    @staticmethod
    def expected_dict():
        return {'OID': 'ODM.SE.BASELINE', 'Name': 'Baseline Visit', 'Repeating': 'No', 'Type': 'Scheduled',
                'Category': 'Pre-treatment', 'FormRef': [{'FormOID': 'ODM.F.VS', 'Mandatory': 'Yes', 'OrderNumber': 1},
                                                  {'FormOID': 'ODM.F.DM', 'Mandatory': 'Yes', 'OrderNumber': 2},
                                                  {'FormOID': 'ODM.F.MH', 'Mandatory': 'Yes', 'OrderNumber': 3}],
                'Description': {'TranslatedText': [{'lang': 'en', '_content': 'this is the first test description'}]},
                'Alias': [{'Context': 'SDTMIG', 'Name': 'VS'}]}



from unittest import TestCase
import odmlib.odm_1_3_2.model as ODM
import datetime
import xml.etree.ElementTree as ET
import os
import odmlib.odm_loader as OL
import odmlib.loader as LD


class TestODM(TestCase):
    def setUp(self) -> None:
        self.odm = self.add_root()
        self.odm_test_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'test_odm_001.xml')
        self.odm_writer_test_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'test_odm_002.xml')
        self.loader = LD.ODMLoader(OL.XMLODMLoader())

    def test_odm_attributes(self):
        self.assertEqual("ODM.MDV.TEST.001", self.odm.FileOID)
        self.assertEqual("Metadata", self.odm.Granularity)

    def test_odm_xml_file(self):
        study = self.add_study()
        self.odm.Study = [study]
        odm_xml = self.odm.to_xml()
        self.write_odm_file(odm_xml)
        found_list = [e.tag for e in odm_xml.getchildren()]
        print(found_list)
        self.assertListEqual(["Study"], [e.tag for e in odm_xml.getchildren()])

    def test_odm_xml_writer(self):
        study = self.add_study()
        self.odm.Study = [study]
        self.odm.write_xml(self.odm_writer_test_file)
        # now test roundtrip by parsing the file
        self.loader.open_odm_document(self.odm_writer_test_file)
        mdv = self.loader.MetaDataVersion()
        self.assertEqual(mdv.Name, "TRACE-XML MDV")
        self.assertEqual(mdv.OID, "MDV.TRACE-XML-ODM-01")

    def test_odm_dict(self):
        study = self.add_study()
        self.odm.Study = [study]
        odm_dict = self.odm.to_dict()
        print(odm_dict)
        self.assertDictEqual(self.expected_dict(), odm_dict)

    def write_odm_file(self, odm_xml):
        tree = ET.ElementTree(odm_xml)
        tree.write(self.odm_test_file, xml_declaration=True)

    def set_datetime(self):
        return datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()

    def add_study(self):
        study_name = ODM.StudyName(_content="ODM XML Test Study Name")
        protocol_name = ODM.ProtocolName(_content="ODM XML Test Study")
        study_description = ODM.StudyDescription(_content="Testing the generation of an ODM XML file")
        gv = ODM.GlobalVariables()
        gv.StudyName = study_name
        gv.StudyDescription = study_description
        gv.ProtocolName = protocol_name
        study = ODM.Study(OID="ODM.STUDY.001")
        study.GlobalVariables = gv
        mdv = self.add_mdv()
        study.MetaDataVersion = [mdv]
        return study

    def add_root(self):
        attrs = self.get_root_attributes()
        root = ODM.ODM(**attrs)
        return root

    def get_root_attributes(self):
        attrs = {"FileOID": "ODM.MDV.TEST.001", "Granularity": "Metadata",
                 "AsOfDateTime": "2020-07-13T00:13:51.309617+00:00",
                 "CreationDateTime": "2020-07-13T00:13:51.309617+00:00", "ODMVersion": "1.3.2", "FileType": "Snapshot",
                 "Originator": "RDS", "SourceSystem": "ODMLib", "SourceSystemVersion": "0.1",
                 "schemaLocation": "http://www.cdisc.org/ns/odm/v1.3 odm1-3-2.xsd"}
        return attrs

    def add_mdv(self):
        attrs = self.set_mdv_attributes()
        mdv = ODM.MetaDataVersion(**attrs)
        mdv.Protocol = self.add_protocol()
        mdv.StudyEventDef = self.add_SED()
        mdv.FormDef = self.add_FD()
        mdv.ItemGroupDef = self.add_IGD()
        mdv.ItemDef = self.add_ITD()
        mdv.CodeList = self.add_CL()
        mdv.ConditionDef = self.add_CD()
        mdv.MethodDef = self.add_MD()
        return mdv

    def add_CD(self):
        tt1 = ODM.TranslatedText(_content="Skip the BRTHMO field when BRTHYR length NE 4", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt1]
        cd = ODM.ConditionDef(OID="ODM.CD.BRTHMO", Name="Skip BRTHMO when no BRTHYR")
        cd.Description = desc
        return [cd]

    def add_MD(self):
        tt1 = ODM.TranslatedText(_content="Concatenation of BRTHYR, BRTHMO, and BRTHDY in ISO 8601 format", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt1]
        md = ODM.MethodDef(OID="ODM.MT.DOB", Name="Create BRTHDTC from date ELEMENTS", Type="Computation")
        md.Description = desc
        return [md]

    def add_CL(self):
        tt1 = ODM.TranslatedText(_content="No", lang="en")
        dc1 = ODM.Decode()
        dc1.TranslatedText = [tt1]
        cli1 = ODM.CodeListItem(CodedValue="N")
        cli1.Decode = dc1
        tt2 = ODM.TranslatedText(_content="Yes", lang="en")
        dc2 = ODM.Decode()
        dc2.TranslatedText = [tt2]
        cli2 = ODM.CodeListItem(CodedValue="Y")
        cli2.Decode = dc2
        cl = ODM.CodeList(OID="ODM.CL.NY_SUB_Y_N", Name="No Yes Response", DataType="text")
        cl.CodeListItem = [cli1, cli2]
        return [cl]


    def add_ITD(self):
        # ItemDef 1
        ttd1 = ODM.TranslatedText(_content="Date of measurements", lang="en")
        ttq1 = ODM.TranslatedText(_content="Date", lang="en")
        desc1 = ODM.Description()
        desc1.TranslatedText = [ttd1]
        q1 = ODM.Question()
        q1.TranslatedText = [ttq1]
        a1 = ODM.Alias(Context="CDASH", Name="VSDAT")
        itd1 = ODM.ItemDef(OID="ODM.IT.VS.VSDAT", Name="Date", DataType="partialDate")
        itd1.Description = desc1
        itd1.Question = q1
        itd1.Alias = [a1]
        # ItemDef 2
        ttd2 = ODM.TranslatedText(_content="Result of the vital signs measurement as originally received or collected.", lang="en")
        ttq2 = ODM.TranslatedText(_content="Diastolic", lang="en")
        desc2 = ODM.Description()
        desc2.TranslatedText = [ttd2]
        q2 = ODM.Question()
        q2.TranslatedText = [ttq2]
        a2a = ODM.Alias(Context="CDASH", Name="BP.DIABP.VSORRES")
        a2b = ODM.Alias(Context="CDASH/SDTM", Name="VSORRES+VSORRESU")
        itd2 = ODM.ItemDef(OID="ODM.IT.VS.BP.VSORRESU", Name="BP Units", DataType="text")
        itd2.Description = desc2
        itd2.Question = q2
        itd2.Alias = [a2a, a2b]
        return [itd1, itd2]

    def add_IGD(self):
        itr1 = ODM.ItemRef(ItemOID="ODM.IT.VS.VSDAT", Mandatory="Yes")
        itr2 = ODM.ItemRef(ItemOID="ODM.IT.VS.BP.DIABP.VSORRES", Mandatory="Yes")
        itr3 = ODM.ItemRef(ItemOID="ODM.IT.VS.BP.SYSBP.VSORRES", Mandatory="Yes")
        igd1 = ODM.ItemGroupDef(OID="ODM.IG.VS", Name="Vital Sign Measurement", Repeating="Yes")
        igd1.ItemRef = [itr1, itr2, itr3]
        igr4 = ODM.ItemRef(ItemOID="ODM.IT.DM.BRTHYR", Mandatory="Yes")
        igr5 = ODM.ItemRef(ItemOID="ODM.IT.DM.SEX", Mandatory="Yes")
        igd2 = ODM.ItemGroupDef(OID="ODM.IG.DM", Name="Demographics", Repeating="No")
        igd2.ItemRef = [igr4, igr5]
        return [igd1, igd2]

    def add_FD(self):
        igr1 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.Common", Mandatory="Yes", OrderNumber=1)
        igr2 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS_GENERAL", Mandatory="Yes", OrderNumber=2)
        igr3 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS", Mandatory="Yes", OrderNumber=3)
        fd1 = ODM.FormDef(OID="ODM.F.VS", Name="Vital Signs", Repeating="No")
        fd1.ItemGroupRef = [igr1, igr2, igr3]
        igr4 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.Common", Mandatory="Yes", OrderNumber=1)
        igr5 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.DM", Mandatory="Yes", OrderNumber=2)
        fd2 = ODM.FormDef(OID="ODM.F.DM", Name="Demographics", Repeating="No")
        fd2.ItemGroupRef = [igr4, igr5]
        return [fd1, fd2]

    def add_SED(self):
        fr1 = ODM.FormRef(FormOID="ODM.F.DM", Mandatory="Yes", OrderNumber=1)
        fr2 = ODM.FormRef(FormOID="ODM.F.VS", Mandatory="Yes", OrderNumber=2)
        fr3 = ODM.FormRef(FormOID="ODM.F.AE", Mandatory="Yes", OrderNumber=3)
        ser1 = ODM.StudyEventDef(OID="BASELINE", Name="Baseline Visit", Repeating="No", Type="Scheduled")
        ser1.FormRef = [fr1, fr2, fr3]
        ser2 = ODM.StudyEventDef(OID="FOLLOW-UP", Name="Follow-up Visit", Repeating="Yes", Type="Scheduled")
        ser2.FormRef = [fr1, fr2, fr3]
        return [ser1, ser2]

    def add_protocol(self):
        p = ODM.Protocol()
        tt = ODM.TranslatedText(_content="Trace-XML Test CDASH File", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt]
        ser1 = ODM.StudyEventRef(StudyEventOID="BASELINE", OrderNumber=1, Mandatory="Yes")
        ser2 = ODM.StudyEventRef(StudyEventOID="FOLLOW-UP", OrderNumber=2, Mandatory="Yes")
        a = ODM.Alias(Context="ClinicalTrials.gov", Name="trace-protocol")
        p.Description = desc
        p.StudyEventRef = [ser1, ser2]
        p.Alias = [a]
        return p

    def set_mdv_attributes(self):
        return {"OID": "MDV.TRACE-XML-ODM-01", "Name": "TRACE-XML MDV", "Description": "Trace-XML Example"}

    def expected_dict(self):
        return {
    "FileOID": "ODM.MDV.TEST.001",
    "Granularity": "Metadata",
    "AsOfDateTime": "2020-07-13T00:13:51.309617+00:00",
    "CreationDateTime": "2020-07-13T00:13:51.309617+00:00",
    "ODMVersion": "1.3.2",
    "FileType": "Snapshot",
    "Originator": "RDS",
    "SourceSystem": "ODMLib",
    "SourceSystemVersion": "0.1",
    "schemaLocation": "http://www.cdisc.org/ns/odm/v1.3 odm1-3-2.xsd",
    "Study": [{
        "OID": "ODM.STUDY.001",
        "GlobalVariables": {
            "StudyName": {"_content": "ODM XML Test Study Name"},
            "StudyDescription": {"_content": "Testing the generation of an ODM XML file"},
            "ProtocolName": {"_content": "ODM XML Test Study"}
        },
        "MetaDataVersion": [{
            "OID": "MDV.TRACE-XML-ODM-01",
            "Name": "TRACE-XML MDV",
            "Description": "Trace-XML Example",
            "Protocol": {
                "Description": {"TranslatedText": [{
                    "_content": "Trace-XML Test CDASH File",
                    "lang": "en"
                }]},
                "StudyEventRef": [
                    {
                        "StudyEventOID": "BASELINE",
                        "OrderNumber": 1,
                        "Mandatory": "Yes"
                    },
                    {
                        "StudyEventOID": "FOLLOW-UP",
                        "OrderNumber": 2,
                        "Mandatory": "Yes"
                    }
                ],
                "Alias": [{
                    "Context": "ClinicalTrials.gov",
                    "Name": "trace-protocol"
                }]
            },
            "StudyEventDef": [
                {
                    "OID": "BASELINE",
                    "Name": "Baseline Visit",
                    "Repeating": "No",
                    "Type": "Scheduled",
                    "FormRef": [
                        {
                            "FormOID": "ODM.F.DM",
                            "Mandatory": "Yes",
                            "OrderNumber": 1
                        },
                        {
                            "FormOID": "ODM.F.VS",
                            "Mandatory": "Yes",
                            "OrderNumber": 2
                        },
                        {
                            "FormOID": "ODM.F.AE",
                            "Mandatory": "Yes",
                            "OrderNumber": 3
                        }
                    ]
                },
                {
                    "OID": "FOLLOW-UP",
                    "Name": "Follow-up Visit",
                    "Repeating": "Yes",
                    "Type": "Scheduled",
                    "FormRef": [
                        {
                            "FormOID": "ODM.F.DM",
                            "Mandatory": "Yes",
                            "OrderNumber": 1
                        },
                        {
                            "FormOID": "ODM.F.VS",
                            "Mandatory": "Yes",
                            "OrderNumber": 2
                        },
                        {
                            "FormOID": "ODM.F.AE",
                            "Mandatory": "Yes",
                            "OrderNumber": 3
                        }
                    ]
                }
            ],
            "FormDef": [
                {
                    "OID": "ODM.F.VS",
                    "Name": "Vital Signs",
                    "Repeating": "No",
                    "ItemGroupRef": [
                        {
                            "ItemGroupOID": "ODM.IG.Common",
                            "Mandatory": "Yes",
                            "OrderNumber": 1
                        },
                        {
                            "ItemGroupOID": "ODM.IG.VS_GENERAL",
                            "Mandatory": "Yes",
                            "OrderNumber": 2
                        },
                        {
                            "ItemGroupOID": "ODM.IG.VS",
                            "Mandatory": "Yes",
                            "OrderNumber": 3
                        }
                    ]
                },
                {
                    "OID": "ODM.F.DM",
                    "Name": "Demographics",
                    "Repeating": "No",
                    "ItemGroupRef": [
                        {
                            "ItemGroupOID": "ODM.IG.Common",
                            "Mandatory": "Yes",
                            "OrderNumber": 1
                        },
                        {
                            "ItemGroupOID": "ODM.IG.DM",
                            "Mandatory": "Yes",
                            "OrderNumber": 2
                        }
                    ]
                }
            ],
            "ItemGroupDef": [
                {
                    "OID": "ODM.IG.VS",
                    "Name": "Vital Sign Measurement",
                    "Repeating": "Yes",
                    "ItemRef": [
                        {
                            "ItemOID": "ODM.IT.VS.VSDAT",
                            "Mandatory": "Yes"
                        },
                        {
                            "ItemOID": "ODM.IT.VS.BP.DIABP.VSORRES",
                            "Mandatory": "Yes"
                        },
                        {
                            "ItemOID": "ODM.IT.VS.BP.SYSBP.VSORRES",
                            "Mandatory": "Yes"
                        }
                    ]
                },
                {
                    "OID": "ODM.IG.DM",
                    "Name": "Demographics",
                    "Repeating": "No",
                    "ItemRef": [
                        {
                            "ItemOID": "ODM.IT.DM.BRTHYR",
                            "Mandatory": "Yes"
                        },
                        {
                            "ItemOID": "ODM.IT.DM.SEX",
                            "Mandatory": "Yes"
                        }
                    ]
                }
            ],
            "ItemDef": [
                {
                    "OID": "ODM.IT.VS.VSDAT",
                    "Name": "Date",
                    "DataType": "partialDate",
                    "Description": {"TranslatedText": [{
                        "_content": "Date of measurements",
                        "lang": "en"
                    }]},
                    "Question": {"TranslatedText": [{
                        "_content": "Date",
                        "lang": "en"
                    }]},
                    "Alias": [{
                        "Context": "CDASH",
                        "Name": "VSDAT"
                    }]
                },
                {
                    "OID": "ODM.IT.VS.BP.VSORRESU",
                    "Name": "BP Units",
                    "DataType": "text",
                    "Description": {"TranslatedText": [{
                        "_content": "Result of the vital signs measurement as originally received or collected.",
                        "lang": "en"
                    }]},
                    "Question": {"TranslatedText": [{
                        "_content": "Diastolic",
                        "lang": "en"
                    }]},
                    "Alias": [
                        {
                            "Context": "CDASH",
                            "Name": "BP.DIABP.VSORRES"
                        },
                        {
                            "Context": "CDASH/SDTM",
                            "Name": "VSORRES+VSORRESU"
                        }
                    ]
                }
            ],
            "CodeList": [{
                "OID": "ODM.CL.NY_SUB_Y_N",
                "Name": "No Yes Response",
                "DataType": "text",
                "CodeListItem": [
                    {
                        "CodedValue": "N",
                        "Decode": {"TranslatedText": [{
                            "_content": "No",
                            "lang": "en"
                        }]}
                    },
                    {
                        "CodedValue": "Y",
                        "Decode": {"TranslatedText": [{
                            "_content": "Yes",
                            "lang": "en"
                        }]}
                    }
                ]
            }],
            "ConditionDef": [{
                "OID": "ODM.CD.BRTHMO",
                "Name": "Skip BRTHMO when no BRTHYR",
                "Description": {"TranslatedText": [{
                    "_content": "Skip the BRTHMO field when BRTHYR length NE 4",
                    "lang": "en"
                }]}
            }],
            "MethodDef": [{
                "OID": "ODM.MT.DOB",
                "Name": "Create BRTHDTC from date ELEMENTS",
                "Type": "Computation",
                "Description": {"TranslatedText": [{
                    "_content": "Concatenation of BRTHYR, BRTHMO, and BRTHDY in ISO 8601 format",
                    "lang": "en"
                }]}
            }]
        }]
    }]
}





class Descriptor:
    def __init__(self, name=None, required=False, element_class=None, valid_values=[], namespace="odm"):
        self.name = name
        self.required = required
        self.element_class = element_class
        self.valid_values = valid_values
        self.namespace = namespace

    def __get__(self, instance, cls):
        if instance is None:
            return self
        elif (self.name not in instance.__dict__) and (self.name != self.__dict__["name"]):
            raise ValueError(f"Missing attribute or element {self.name} in {cls.__name__}")
        else:
            if self.name not in instance.__dict__:
                if isinstance(self, list):
                    self.__set__(instance, [])
                else:
                    if self.element_class is None:
                        instance.__dict__[self.name] = self.element_class
                    else:
                        try:
                            instance.__dict__[self.name] = self.element_class()
                        except ValueError:
                            instance.__dict__[self.name] = None
            return instance.__dict__[self.name]

    def __set__(self, instance, value):
        instance.__dict__[self.name] = value

    def __delete__(self, instance):
        del instance.__dict__[self.name]
